'''Enable instrumentationon all supported modules.''' # pylint: disable=R0401
import logging
from tyke.agent import Agent
from tyke.env_var_settings import get_env_value

# Initialize logger
logger = logging.getLogger(__name__)  # pylint: disable=C0103

skip_modules = get_env_value('SKIP_MODULES')
if skip_modules:
    logger.debug("[env] Loaded SKIP_MODULES from env")
    if len(skip_modules) > 0:
        skip_modules = skip_modules.replace(' ', '')
        skip_modules = skip_modules.split(',')
else:
    skip_modules = []

# Create Tyke agent
agent = Agent()
agent.instrument(None, skip_modules, auto_instrument=True)



'''this module acts as a driver for instrumentation definitions + application'''
import logging

FLASK_KEY = 'flask'
DJANGO_KEY = 'django'
FAST_API_KEY = 'fastapi'
GRPC_SERVER_KEY = 'grpc:server'
GRPC_CLIENT_KEY = 'grpc:client'
POSTGRESQL_KEY = 'postgresql'
MYSQL_KEY = 'mysql'
REQUESTS_KEY = 'requests'
AIOHTTP_CLIENT_KEY = 'aiohttp:client'
LAMBDA = 'lambda'
BOTO = 'boto'
BOTOCORE = 'botocore'

SUPPORTED_LIBRARIES = [
    FLASK_KEY, DJANGO_KEY, FAST_API_KEY,
    GRPC_SERVER_KEY, GRPC_CLIENT_KEY,
    POSTGRESQL_KEY, MYSQL_KEY,
    REQUESTS_KEY, AIOHTTP_CLIENT_KEY,
    LAMBDA, BOTO, BOTOCORE
]

# map of library_key => instrumentation wrapper instance
_INSTRUMENTATION_STATE = {}

logger = logging.getLogger(__name__)

def _uninstrument_all():
    for key in _INSTRUMENTATION_STATE:
        logger.debug("Uninstrumenting %s", key)
        _INSTRUMENTATION_STATE[key].uninstrument()

    _INSTRUMENTATION_STATE.clear()

def is_already_instrumented(library_key):
    """check if an instrumentation wrapper is already registered"""
    return library_key in _INSTRUMENTATION_STATE.keys()


def _mark_as_instrumented(library_key, wrapper_instance):
    """mark an instrumentation wrapper as registered"""
    _INSTRUMENTATION_STATE[library_key] = wrapper_instance


def get_instrumentation_wrapper(library_key): # pylint:disable=R0912
    """load an initialize an instrumentation wrapper"""
    if is_already_instrumented(library_key):
        return None
    try:
        wrapper_instance = None
        if DJANGO_KEY == library_key:
            from tyke.agent.instrumentation.django import DjangoInstrumentationWrapper  #pylint:disable=C0415
            wrapper_instance = DjangoInstrumentationWrapper()
        elif FLASK_KEY == library_key:
            from tyke.agent.instrumentation.flask import FlaskInstrumentorWrapper #pylint:disable=C0415
            wrapper_instance = FlaskInstrumentorWrapper()
        elif FAST_API_KEY == library_key:
            from tyke.agent.instrumentation.fast_api import FastAPIInstrumentorWrapper #pylint:disable=C0415
            wrapper_instance = FastAPIInstrumentorWrapper()
        elif GRPC_SERVER_KEY == library_key:
            from tyke.agent.instrumentation.grpc import GrpcInstrumentorServerWrapper #pylint:disable=C0415
            wrapper_instance = GrpcInstrumentorServerWrapper()
        elif GRPC_CLIENT_KEY == library_key:
            from tyke.agent.instrumentation.grpc import GrpcInstrumentorClientWrapper #pylint:disable=C0415
            wrapper_instance = GrpcInstrumentorClientWrapper()
        elif POSTGRESQL_KEY == library_key:
            from tyke.agent.instrumentation.postgresql import PostgreSQLInstrumentorWrapper #pylint:disable=C0415
            wrapper_instance =  PostgreSQLInstrumentorWrapper()
        elif MYSQL_KEY == library_key:
            from tyke.agent.instrumentation.mysql import MySQLInstrumentorWrapper #pylint:disable=C0415
            wrapper_instance = MySQLInstrumentorWrapper()
        elif REQUESTS_KEY == library_key:
            from tyke.agent.instrumentation.requests import RequestsInstrumentorWrapper #pylint:disable=C0415
            wrapper_instance = RequestsInstrumentorWrapper()
        elif AIOHTTP_CLIENT_KEY == library_key:
            from tyke.agent.instrumentation.aiohttp import AioHttpClientInstrumentorWrapper #pylint:disable=C0415
            wrapper_instance = AioHttpClientInstrumentorWrapper()
        elif LAMBDA == library_key:
            from tyke.agent.instrumentation.aws_lambda import AwsLambdaInstrumentorWrapper #pylint:disable=C0415
            wrapper_instance = AwsLambdaInstrumentorWrapper()
        elif BOTO == library_key:
            from tyke.agent.instrumentation.boto import BotoInstrumentationWrapper #pylint:disable=C0415
            wrapper_instance = BotoInstrumentationWrapper()
        elif BOTOCORE == library_key:
            from tyke.agent.instrumentation.botocore import BotocoreInstrumentationWrapper #pylint:disable=C0415
            wrapper_instance = BotocoreInstrumentationWrapper()
        else:
            return None

        _mark_as_instrumented(library_key, wrapper_instance)
        return wrapper_instance
    except Exception as _err: # pylint:disable=W0703
        logger.debug("Error while attempting to load instrumentation wrapper for %s", library_key)
        return None



'''Tyke instrumentation logic for aiohttp-client'''
import logging
import traceback
import types
import typing
from collections import deque
import asyncio
import aiohttp
import wrapt
from opentelemetry import context as context_api
from opentelemetry import trace
from opentelemetry.instrumentation.aiohttp_client.version import __version__
from opentelemetry.trace import TracerProvider, get_tracer
from opentelemetry.instrumentation.aiohttp_client import AioHttpClientInstrumentor
from tyke.agent.instrumentation import BaseInstrumentorWrapper

# Initialize logger with local module name
logger = logging.getLogger(__name__)  # pylint: disable=C0103

# Max time to wait for response data to be read.
MAX_WAIT_TIME = 0.1 # seconds

# aiohttp-client instrumentation module wrapper class
class AioHttpClientInstrumentorWrapper(AioHttpClientInstrumentor, BaseInstrumentorWrapper):
    """Tyke wrapper class around OpenTelemetry AioHttpClient Instrumentor class"""
    # Constructor
    def __init__(self):
        '''Constructor'''
        logger.debug('Entering AioHttpClientInstrumentor.__init__().')
        super().__init__()

    def _instrument(self, **kwargs) -> None:
        '''Enable instrumentation.'''
        logger.debug(
            'Entering AioHttpClientInstrumentorWrapper._instrument().')
        # Initialize OTel instrumentor
        super()._instrument(
            tracer_provider=kwargs.get("tracer_provider"),
            url_filter=kwargs.get("url_filter"),
            span_name=kwargs.get('span_name')
        )
        # Initialize tyke instrumentor
        _instrument(
            tracer_provider=kwargs.get("tracer_provider"),
            url_filter=kwargs.get("url_filter"),
            span_name=kwargs.get("span_name"),
            aiohttp_client_wrapper=self
        )

# aliases for type definitions
_UrlFilterT = typing.Optional[typing.Callable[[  # pylint: disable=unsubscriptable-object
    str], str]]
_SpanNameT = typing.Optional[  # pylint: disable=unsubscriptable-object
    typing.Union[typing.Callable[[aiohttp.TraceRequestStartParams],  # pylint: disable=unsubscriptable-object
                                 str], str]
]

# build an aiohttp trace config


def create_trace_config(
        url_filter: _UrlFilterT = None,
        span_name: _SpanNameT = None,
        tracer_provider: TracerProvider = None,
        aiohttp_client_wrapper: AioHttpClientInstrumentorWrapper = None
) -> aiohttp.TraceConfig:
    '''Build an aiohttp-client trace config for use with Tyke'''
    tracer = get_tracer(__name__, __version__, tracer_provider)

    # https://docs.aiohttp.org/en/stable/tracing_reference.html describes
    # the events/signals and handlers that can be registered
    # This runs after each chunk of request data is sent
    async def on_request_chunk_sent(
            unused_session: aiohttp.ClientSession,
            trace_config_ctx: types.SimpleNamespace,
            params: aiohttp.TraceRequestChunkSentParams
    ):
        logger.debug('Entering tyke on_request_chunk_sent().')
        if hasattr(params, 'chunk') and params.chunk is not None:
            decoded_chunk = params.chunk.decode()
            logger.debug('request chunk: %s', decoded_chunk)
            trace_config_ctx.request_body += decoded_chunk

    # This runs after an exception occurs
    async def on_request_exception( # pylint: disable=W0613
            unused_session: aiohttp.ClientSession,
            trace_config_ctx: types.SimpleNamespace,
            params: aiohttp.TraceRequestExceptionParams,
    ):
        logger.debug('Entering on_request_exception().')


    # This runs after the request
    async def on_request_end(
            unused_session: aiohttp.ClientSession,
            trace_config_ctx: types.SimpleNamespace,
            params: aiohttp.TraceRequestEndParams,
    ) -> None:
        logger.debug('Entering tyke on_request_end().')
        # utf8_decoder = codecs.getincrementaldecoder('utf-8')
        response_body = b''
        if hasattr(params.response, 'content') \
          and params.response.content is not None:
            content_stream = params.response.content
            # A temporary dual end queue to copy data into
            # and use to reset the stream.
            tmp_deque = deque()
            # Read all the data in the response buffer. This
            # will block until it receives the expected number of bytes
            # or the connection times out and move on with already read data.
            logger.debug('Reading data---->')
            try:
                while not content_stream.at_eof():
                    response_chunk = b''
                    response_chunk = await asyncio.wait_for(
                       content_stream.read(),
                       MAX_WAIT_TIME
                    )
                    tmp_deque.append(response_chunk)
                    response_body += response_chunk
            except asyncio.TimeoutError as err: # pylint: disable=W0703
                logger.error('No data to display, exception=%s, stacktrace=%s',
                             err,
                             traceback.print_exc())
            finally:
                logger.debug('response_body: %s', str(response_body))
                # Reset response.content_stream
                content_stream._cursor = 0 # pylint: disable=W0212
                content_stream._buffer = tmp_deque # pylint: disable=W0212
        request_body = ''
        if hasattr(trace_config_ctx, 'request_body') and trace_config_ctx.request_body is not None:
            request_body = trace_config_ctx.request_body
        span = trace.get_current_span()
        # Add headers & body to span
        if span.is_recording():
            aiohttp_client_wrapper.generic_request_handler(
                params.headers, request_body, span)
            aiohttp_client_wrapper.generic_response_handler(
                params.response.headers, response_body, span)
        trace_config_ctx.end_callback_called = True
        trace_config_ctx.span = span

    def _trace_config_ctx_factory(**kwargs):
        kwargs.setdefault("trace_request_ctx", {})
        return types.SimpleNamespace(
            span_name=span_name,\
            tracer=tracer, \
            url_filter=url_filter, \
            **kwargs, \
            request_body='', \
        )

    trace_config = aiohttp.TraceConfig(
        trace_config_ctx_factory=_trace_config_ctx_factory
    )

    trace_config.on_request_chunk_sent.append(on_request_chunk_sent)
    trace_config.on_request_end.append(on_request_end)
    trace_config.on_request_exception.append(on_request_exception)

    return trace_config

def _instrument(
        tracer_provider: TracerProvider = None,
        url_filter: _UrlFilterT = None,
        span_name: _SpanNameT = None,
        aiohttp_client_wrapper: AioHttpClientInstrumentorWrapper = None
) -> None:
    '''Setup details of trace config context'''
    def instrumented_init(wrapped, instance, args, kwargs) -> None:  # pylint: disable=W0613
        if context_api.get_value("suppress_instrumentation"):
            return wrapped(*args, **kwargs)

        trace_configs = list(kwargs.get("trace_configs") or ())

        trace_config = create_trace_config(
            url_filter=url_filter,
            span_name=span_name,
            tracer_provider=tracer_provider,
            aiohttp_client_wrapper=aiohttp_client_wrapper
        )
        trace_configs.append(trace_config)

        kwargs["trace_configs"] = trace_configs
        return wrapped(*args, **kwargs)

    wrapt.wrap_function_wrapper(
        aiohttp.ClientSession, "__init__", instrumented_init
    )



'''Initialize all the components using configuration from AgentConfig'''
import sys
import os
import traceback
import logging
from typing import Union

from opentelemetry import trace
from opentelemetry.sdk.trace import TracerProvider, export
from opentelemetry.sdk.trace.export import ConsoleSpanExporter, SimpleSpanProcessor
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter as OTLPGrpcSpanExporter
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter as OTLPHttpSpanExporter
from opentelemetry.exporter.zipkin.proto.http import ZipkinExporter
from opentelemetry.trace import ProxyTracerProvider
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from tyke.agent import constants
from tyke.agent.config import config_pb2, AgentConfig

# Initialize logger
logger = logging.getLogger(__name__)  # pylint: disable=C0103


class AgentInit:  # pylint: disable=R0902,R0903
    '''Initialize all the OTel components using configuration from AgentConfig'''

    def __init__(self, agent_config: AgentConfig):
        '''constructor'''
        logger.debug('Initializing AgentInit object.')
        self._config = agent_config

        # Only available in python > 3.7
        # this does prevent user from having to add post fork hooks to their
        # web server config
        if hasattr(os, 'register_at_fork'):
            logger.info('Registering after_in_child handler.')
            os.register_at_fork(after_in_child=self.post_fork)  # pylint:disable=E1101

        try:
            self.apply_config(None)

        except Exception as err:  # pylint: disable=W0703
            logger.error('Failed to initialize tracer: exception=%s, stacktrace=%s',
                         err,
                         traceback.format_exc())
            raise sys.exc_info()[0]

    def post_fork(self):
        """Used to reinitialize exporter & processors in separate worker processes"""
        self.apply_config(None)  # pylint:disable=W0212

    def apply_config(self, agent_config: (Union[None, AgentConfig])):
        """Initialize various aspects of the agent based on the most recent config"""
        if agent_config:
            self._config = agent_config

        self.init_trace_provider()
        self.init_propagation()

        if self._config.use_console_span_exporter():
            self.set_console_span_processor()
        else:
            self.init_exporter()


    def init_trace_provider(self) -> None:
        '''Initialize trace provider and set resource attributes.'''
        if isinstance(trace.get_tracer_provider(), ProxyTracerProvider):
            logger.debug("no configured trace provider detected, adding one")
            resource_attributes = {
                "service.name": self._config.agent_config.service_name,
                "service.instance.id": os.getpid(),
                "telemetry.sdk.version": constants.TELEMETRY_SDK_VERSION,
                "telemetry.sdk.name": constants.TELEMETRY_SDK_NAME,
                "telemetry.sdk.language": constants.TELEMETRY_SDK_LANGUAGE
            }
            if self._config.agent_config.resource_attributes:
                logger.debug(
                    'Custom attributes found. Adding to resource attributes dict.')
                resource_attributes.update(
                    self._config.agent_config.resource_attributes)
            tracer_provider = TracerProvider(
                resource=Resource.create(resource_attributes)
            )
            trace.set_tracer_provider(tracer_provider)
        else:
            logger.debug("tracer provider already configured, "
                         "skipping trace provider configuration")

    def init_exporter(self) -> None:
        """Initialize exporter"""
        reporter_type = self._config.agent_config.reporting.trace_reporter_type
        exporter = self._init_exporter(reporter_type)
        if exporter is None:
            logger.warning("Unable to initialize exporter")
            return

        span_processor = BatchSpanProcessor(exporter)
        trace.get_tracer_provider().add_span_processor(span_processor)

    def init_propagation(self) -> None:
        '''Initialize requested context propagation protocols.'''
        propagator_list = []
        for prop_format in self._config.agent_config.propagation_formats:
            if prop_format == config_pb2.PropagationFormat.TRACECONTEXT:
                from opentelemetry.trace.propagation.tracecontext \
                  import TraceContextTextMapPropagator # pylint: disable=C0415
                propagator_list += [ TraceContextTextMapPropagator() ]
                logger.debug('Adding TRACECONTEXT trace propagator to list.')
            if prop_format == config_pb2.PropagationFormat.B3:
                from opentelemetry.propagators.b3 import B3MultiFormat  # pylint: disable=C0415
                propagator_list += [ B3MultiFormat() ]
                logger.debug('Adding B3 trace propagator to list.')

        if len(propagator_list) == 0:
            logger.debug('No propagators have been initialized.')

        logger.debug('propagator_list: %s', str(propagator_list))
        from opentelemetry.propagate import set_global_textmap # pylint: disable=C0415
        from opentelemetry.propagators.composite import CompositePropagator # pylint: disable=C0415
        composite_propagators = CompositePropagator(propagator_list)
        set_global_textmap(composite_propagators)

    def _set_wrapper_fields(self, wrapper):
        data_cap = self._config.agent_config.data_capture
        wrapper.set_process_request_headers(data_cap.http_headers.request)
        wrapper.set_process_request_body(data_cap.http_body.request)
        wrapper.set_process_response_headers(data_cap.http_headers.response)
        wrapper.set_process_response_body(data_cap.http_body.response)
        wrapper.set_body_max_size(data_cap.body_max_size_bytes)

    def init_library_instrumentation(self, instrumentation_name, wrapper_instance):
        """used to configure instrumentation wrapper settings + apply instrumentation"""
        logger.debug("Attempting to initialize %s instrumentation", instrumentation_name)
        try:
            self._set_wrapper_fields(wrapper_instance)
            wrapper_instance.instrument()
        except Exception as err: # pylint: disable=W0703
            logger.debug(constants.INST_WRAP_EXCEPTION_MSSG,
                         instrumentation_name,
                         err,
                         traceback.format_exc())

    def register_processor(self, processor) -> None:  # pylint:disable=R0201
        '''Register additional span exporter + processor'''
        logger.debug('Entering AgentInit.register_processor().')
        trace.get_tracer_provider().add_span_processor(processor)

    def set_console_span_processor(self) -> None:
        '''Register the console span processor for debugging purposes.'''
        logger.debug('Entering AgentInit.setConsoleSpanProcessor().')
        console_span_exporter = ConsoleSpanExporter(
            service_name=self._config.agent_config.service_name)
        simple_export_span_processor = SimpleSpanProcessor(
            console_span_exporter)
        trace.get_tracer_provider().add_span_processor(simple_export_span_processor)

    def _init_exporter(self, trace_reporter_type):
        exporter_type = ''
        exporter = None
        try:
            if trace_reporter_type == config_pb2.TraceReporterType.ZIPKIN:
                exporter_type = 'zipkin'
                exporter = ZipkinExporter(
                    endpoint=self._config.agent_config.reporting.endpoint
                )
            elif trace_reporter_type == config_pb2.TraceReporterType.OTLP:
                exporter_type = 'otlp'
                exporter = OTLPGrpcSpanExporter(endpoint=self._config.agent_config.reporting.endpoint,
                                            insecure= not self._config.agent_config.reporting.secure)
            elif trace_reporter_type == config_pb2.TraceReporterType.OTLP_HTTP:
                exporter_type = 'otlp_http'
                exporter = OTLPHttpSpanExporter(endpoint=self._config.agent_config.reporting.endpoint)

            if exporter:
                logger.info('Initialized %s exporter reporting to `%s`',
                            exporter_type,
                            self._config.agent_config.reporting.endpoint)
            else:
                logger.error("Unknown exporter type `%s`", trace_reporter_type)

            return exporter
        except Exception as err:  # pylint: disable=W0703
            logger.error('Failed to initialize %s exporter: exception=%s, stacktrace=%s',
                         exporter_type,
                         err,
                         traceback.format_exc())
            return None



'''Tyke flask instrumentor module wrapper.''' # pylint: disable=R0401
import sys
import os.path
import logging
import inspect
import traceback
import json
import flask
from opentelemetry.instrumentation.flask import (
    _InstrumentedFlask,
    FlaskInstrumentor,
    _ENVIRON_SPAN_KEY,
)
from werkzeug.exceptions import Forbidden

from tyke.agent import constants  # pylint: disable=R0801
from tyke.agent.filter.registry import Registry, TYPE_HTTP
from tyke.agent.instrumentation import BaseInstrumentorWrapper

from tyke.agent.config import AgentConfig

# Initialize logger
logger = logging.getLogger(__name__)  # pylint: disable=C0103

# Per request pre-handler
def _tyke_before_request(flask_wrapper):
    '''This function is invoked by flask to set the handler'''
    def tyke_before_request() -> None:
        '''Tyke before_request() method'''
        logger.debug('Entering _tyke_before_request().')
        try:
            # Read span from flask "environment". The global flask.request
            # object keeps track of which request belong to the currently
            # active thread. See
            #   https://flask.palletsprojects.com/en/1.1.x/api/#flask.request
            span = flask.request.environ.get(_ENVIRON_SPAN_KEY)
            # Pull request headers
            # for now, assuming single threaded mode (multiple python processes)
            request_headers = flask.request.headers
            # Pull message body
            request_body = flask.request.data       # same

            span.update_name(str(flask.request.method) + ' ' + str(flask.request.url_rule))

            # Call base request handler
            flask_wrapper.generic_request_handler(request_headers, request_body, span)

            block_result = Registry().apply_filters(span,
                                                    flask.request.url,
                                                    flask.request.headers,
                                                    flask.request.data,
                                                    TYPE_HTTP)
            if block_result:
                logger.debug('should block evaluated to true, aborting with 403')
                flask.abort(403)

        except Forbidden as forbidden_error:
            raise forbidden_error

        except Exception as err:  # pylint: disable=W0703
            logger.error(constants.INST_RUNTIME_EXCEPTION_MSSG,
                         'flask before_request handler',
                         err,
                         traceback.format_exc())
            # Not rethrowing to avoid causing runtime errors for Flask.
    return tyke_before_request

# Per request post-handler
def _tyke_after_request(flask_wrapper) -> flask.wrappers.Response:
    '''This function is invoked by flask to set the handler'''
    def tyke_after_request(response):
        '''Tyke after_request method.'''
        try:
            logger.debug('Entering _tyke_after_request().')
            # Read span from flask "environment"
            span = flask.request.environ.get(_ENVIRON_SPAN_KEY)
            # Pull response headers
            response_headers = response.headers

            response_body = ""
            # dont extract response content if body is a file
            if not response.direct_passthrough:
                response_body = response.data

            # Call base response handler
            flask_wrapper.generic_response_handler(
                response_headers, response_body, span)
            return response
        except Exception as err:  # pylint: disable=W0703
            logger.error(constants.INST_RUNTIME_EXCEPTION_MSSG,
                         'flask after_request handler',
                         err,
                         traceback.format_exc())
            # Not rethrowing to avoid causing runtime errors for Flask.
            return response

    return tyke_after_request

class _TykeInstrumentedFlask(_InstrumentedFlask, BaseInstrumentorWrapper):
    """Tyke Wrapper class around OTel _InstrumentedFlask. This replaces
    the flask.Flask class definition."""

    def __init__(self, *args, **kwargs):
        _InstrumentedFlask.__init__(self, *args, **kwargs)
        BaseInstrumentorWrapper.__init__(self)
        self.before_request(_tyke_before_request(self))
        self.after_request(_tyke_after_request(self))
        config: AgentConfig = AgentConfig()
        self.set_process_request_headers(config.agent_config.data_capture.http_headers.request)
        self.set_process_request_body(config.agent_config.data_capture.http_body.request)
        self.set_process_response_headers(config.agent_config.data_capture.http_headers.response)
        self.set_process_response_body(config.agent_config.data_capture.http_body.response)
        self.set_body_max_size(config.agent_config.data_capture.body_max_size_bytes)

# Main Flask Instrumentor Wrapper class.
class FlaskInstrumentorWrapper(FlaskInstrumentor, BaseInstrumentorWrapper):
    '''Tyke wrapper around OTel Flask instrumentor class'''

    def __init__(self):
        logger.debug('Entering FlaskInstrumentorWrapper constructor.')
        super().__init__()
        self._app = None

    def with_app(self, app=None):
        """when instrumenting via code we need to instrument
        the app directly, this is conditionally called from agent.instrument"""
        self._app = app

    def instrument(self, **kwargs):
        if self._app:
            # code based instrumentation
            before_hook = _tyke_before_request(self)
            after_hook = _tyke_after_request(self)
            FlaskInstrumentorWrapper.instrument_app(self._app)
            self._app.before_request(before_hook)
            self._app.after_request(after_hook)
        else:
            # auto instrumentation
            super().instrument()


    def _instrument(self, **kwargs):
        '''Override OTel method that sets up global flask instrumentation'''
        self._original_flask = flask.Flask # pylint: disable = W0201
        name_callback = kwargs.get("name_callback")
        tracer_provider = kwargs.get("tracer_provider")
        if callable(name_callback):
            _TykeInstrumentedFlask.name_callback = name_callback
        _TykeInstrumentedFlask._tracer_provider = tracer_provider # pylint: disable=W0212
        flask.Flask = _TykeInstrumentedFlask

    # Initialize instrumentation wrapper
    @staticmethod
    def instrument_app(app, request_hook=None, response_hook=None, tracer_provider=None, excluded_urls=None):  # pylint:disable=W0221,W0613
        '''Initialize instrumentation'''
        logger.debug('Entering FlaskInstrumentorWrapper.instument_app().')
        try:

            # Call parent class's initialization
            FlaskInstrumentor.instrument_app(app, request_hook, response_hook)

        except Exception as err:  # pylint: disable=W0703
            logger.error("""An error occurred initializing flask otel
                            instrumentor: exception=%s, stacktrace=%s""",
                         err,
                         traceback.format_exc())
            raise err

    # Teardown instrumentation wrapper
    @staticmethod
    def uninstrument_app(app) -> None:
        '''Disable instrumentation'''
        logger.debug('Entering FlaskInstrumentorWrapper.uninstrument_app()')
        try:
            # Call parent's teardown logic
            super().uninstrument_app(app)  # pylint: disable=E1101

        except Exception as err:  # pylint: disable=W0703
            logger.error("""An error occurred while shutting down flask otel
                         instrumentor: exception=%s, stacktrace=%s""",
                         err,
                         traceback.format_exc())
            raise err

    # retrieve flask app
    def get_app(self) -> flask.Flask:
        '''Return the flask app object.'''
        return self._app



'''Tyke wrapper around OTel MySQL Instrumentor''' # pylint: disable=R0801
import logging
from opentelemetry.instrumentation.mysql import MySQLInstrumentor
from tyke.agent.instrumentation import BaseInstrumentorWrapper

# Initialize logger with local module name
logger = logging.getLogger(__name__) # pylint: disable=C0103

# The main entry point
class MySQLInstrumentorWrapper(MySQLInstrumentor, BaseInstrumentorWrapper):
    '''Tyke wrapper around OTel MySQL Instrumentor class'''



"""
we need a post-settings configured hook for django; however that isn't available, but we can wrap
the application getters of wsgi/asgi to run our instrumentation post app init
"""
import logging

from tyke.agent.instrumentation.instrumentation_definitions import DJANGO_KEY

logger = logging.getLogger(__name__)

def add_django_auto_instr_wrappers(agent_init, instrumentation_wrapper):
    """attempt to add wsgi/asgi wrappers if auto instrumentation is used"""
    logger.debug('attempting to add wsgi + asgi wrappers for autoinstrumentation')
    try:
        add_wsgi_wrapper(agent_init, instrumentation_wrapper)
    except:  # pylint:disable=W0702
        logger.error('django.core.wsgi.get_wsgi_application wrapping failed, django may not be instrumented')

    try:
        add_asgi_wrapper(agent_init, instrumentation_wrapper)
    except:  # pylint:disable=W0702
        logger.error('django.core.asgi.get_asgi_application wrapping failed, django may not be instrumented')


def add_wsgi_wrapper(agent_init, instrumentation_wrapper):
    """wrap the asgi wrapper get_asgi_application fn"""
    try:
        from django.core import wsgi  # pylint:disable=C0415
    except:  # pylint:disable=W0702
        logger.error('failed to import django.core.wsgi - wsgi django app may not be instrumented')
        return

    original_get_wsgi_application = wsgi.get_wsgi_application
    wsgi.get_wsgi_application = lambda: apply_wrapper_wsgi_get_app_fn(original_get_wsgi_application, agent_init,
                                                                 instrumentation_wrapper, wsgi)


def add_asgi_wrapper(agent_init, instrumentation_wrapper):
    """wrap the asgi wrapper get_asgi_application fn"""
    try:
        from django.core import asgi  # pylint:disable=C0415
    except:  # pylint:disable=W0702
        logger.error('failed to import django.core.asgi - asgi django app may not be instrumented')
        return

    original_get_asgi_application = asgi.get_asgi_application
    asgi.get_asgi_application = lambda: apply_wrapper_asgi_get_app_fn(original_get_asgi_application, agent_init,
                                                                     instrumentation_wrapper, asgi)

def tyke_get_application(original_fn, agent_init,
                          instrumentation_wrapper, module, method):
    """We need to run instrumentation first to inject middleware before we call the original function
    otherwise middleware stack changes are not applied"""
    try:
        agent_init.register_library(DJANGO_KEY, instrumentation_wrapper)
    except:  # pylint:disable=W0702
        logger.error('registering django instrumentation in %s patch failed '
                     '- continuing without instrumenting django', method)
    app = original_fn()
    module.__dict__[method] = original_fn

    return app

def apply_wrapper_wsgi_get_app_fn(original_fn, agent_init, instrumentation_wrapper, module):
    """wrapper function that calls original app getter and then registers django instrumentation"""
    return tyke_get_application(original_fn, agent_init, instrumentation_wrapper, module, 'get_wsgi_application')

def apply_wrapper_asgi_get_app_fn(original_fn, agent_init, instrumentation_wrapper, module):
    """wrapper function that calls original app getter and then registers django instrumentation"""
    return tyke_get_application(original_fn, agent_init, instrumentation_wrapper, module, 'get_asgi_application')



"""Used for holding filters to apply to ingress traffic"""

from typing import Union, Type
from opentelemetry.trace import Span

from tyke.agent.filter import Filter

TYPE_HTTP = 'http'
TYPE_RPC = 'rpc'

class Registry:
    """
    Registry is used to register and apply request filters
    """
    def __new__(cls):
        if not hasattr(cls, 'instance'):
            cls.instance = super(Registry, cls).__new__(cls)
        return cls.instance

    def __init__(self):
        if not hasattr(self, '_initialized'):
            self._initialized = True
            self.filters = []

    def register(self, filter_class: Type[Filter]):
        '''Register a filter to apply on ingress traffic'''
        instance = filter_class()
        self.filters.append(instance)

    def apply_filters(self, span: Span, url: Union[str, None], headers: dict, body, request_type) -> bool: # pylint:disable=R0913
        '''Apply all registered filters'''
        if url or headers:
            for filter_instance in self.filters:
                if filter_instance.evaluate_url_and_headers(span, url, headers, request_type):
                    return True

        if body:
            for filter_instance in self.filters:
                if filter_instance.evaluate_body(span, body, headers, request_type):
                    return True

        return False



# Based upon the OTel autoinstrumentation feature
'''This module implements a CLI command that be used to
autoinstrument existing pythong programs that use supported
modules.'''
import argparse
from logging import getLogger
from os import environ, execl, getcwd
from os.path import abspath, dirname, pathsep
from shutil import which

logger = getLogger(__file__)

def parse_args():
    '''Parse CLI arguments.'''
    parser = argparse.ArgumentParser(
        description="""
        Tyke-instrument automatically instruments a Python
        program and runs the program
        """
    )

    parser.add_argument("command", help="Your Python application.")

    parser.add_argument(
        "command_args",
        help="Arguments for your application.",
        nargs=argparse.REMAINDER,
    )

    return parser.parse_args()

def update_python_path() -> None:
    '''Retrieve existing PYTHONPATH'''
    python_path = environ.get("PYTHONPATH")

    # Split the paths
    if not python_path:
        python_path = []
    else:
        python_path = python_path.split(pathsep)

    # Get the current working directory
    cwd_path = getcwd()

    # If this directory is already in python_path, remove it.
    python_path = [path for path in python_path if path != cwd_path]

    # Add CWD to the front.
    python_path.insert(0, cwd_path)

    # What is the directory containing this python file?
    filedir_path = dirname(abspath(__file__))

    # If this directory is already in python_path, remove it.
    python_path = [path for path in python_path if path != filedir_path]

    # Add this directory to the front
    python_path.insert(0, filedir_path)

    # Reset PYTHONPATH environment variable
    environ["PYTHONPATH"] = pathsep.join(python_path)

def run() -> None:
    '''Tyke-instrument Entry point'''
    args = parse_args()

    # update PYTHONPATH env var
    update_python_path()

    # Get full path to the command that was passed in as an
    # argument
    executable = which(args.command)

    # Execute the app
    execl(executable, executable, *args.command_args)

if __name__ == '__main__':
    run()



'''File config loader'''

import os
import logging
import traceback
import yaml


logger = logging.getLogger(__name__)  # pylint: disable=C0103


def load_config_from_file(filepath):
    """
    Returns the config loaded from a provided config file
    """
    logger.debug(
        'TYKE_CONFIG_FILE is set %s. Attempting to load the config file', filepath)
    try:
        path = os.path.abspath(filepath)

        file = open(path, 'r') # pylint: disable=R1732
        from_file_config = yaml.safe_load(file)
        file.close()

        logger.debug('Successfully load config from %s', path)

        return from_file_config
    except Exception as err:  # pylint: disable=W0703
        logger.error('Failed to load TYKE_CONFIG_FILE: exception=%s, stacktrace=%s',
                     err,
                     traceback.format_exc())
        logger.info('Loading default configuration.')
        return {}



"""
we need a post-settings configured hook for django; however that isn't available, but we can wrap
the application getters of wsgi/asgi to run our instrumentation post app init
"""
import logging

from tyke.agent.instrumentation.instrumentation_definitions import FAST_API_KEY

logger = logging.getLogger(__name__)
_PATCHED_SETUP = False
_ORIGINAL_SETUP = None

def add_fast_api_auto_instr_wrappers(agent_init, instrumentation_wrapper):
    """wrap the fast api initializer"""
    logger.debug('attempting to add fast api initializer wrapper for autoinstrumentation')
    try:
        global _PATCHED_SETUP  # pylint:disable=W0603
        if not _PATCHED_SETUP:  # pylint:disable=W0603
            add_setup_wrapper(agent_init, instrumentation_wrapper)
            _PATCHED_SETUP = True
    except:  # pylint:disable=W0702
        logger.error('django.core.wsgi.get_wsgi_application wrapping failed, django may not be instrumented')


def add_setup_wrapper(agent_init, instrumentation_wrapper):
    """wrap the asgi wrapper get_asgi_application fn"""
    try:
        import fastapi  # pylint:disable=C0415
    except:  # pylint:disable=W0702
        logger.error('failed to import fastapi -fastapi app may not be instrumented')
        return
    global _ORIGINAL_SETUP  # pylint:disable=W0603
    _ORIGINAL_SETUP = fastapi.FastAPI.setup
    fastapi.FastAPI.setup = lambda fastapi_app: apply_wrapper_setup_fn(_ORIGINAL_SETUP, fastapi_app,
                                                                       agent_init,
                                                                       instrumentation_wrapper)

def unwrap():
    """Used to unwrap the fastapi setup function"""
    try:
        import fastapi  # pylint:disable=C0415
    except:  # pylint:disable=W0702
        logger.error('failed to import fastapi -fastapi app may not be instrumented')
        return
    fastapi.FastAPI.setup = _ORIGINAL_SETUP
    global _PATCHED_SETUP # pylint:disable=W0603
    _PATCHED_SETUP = False


def apply_wrapper_setup_fn(original_fn, fastapi_app, agent_init, instrumentation_wrapper):
    """wrapper function that calls original app getter and then registers django instrumentation"""

    # We need to run instrumentation first to inject middleware before we call the original function
    # otherwise middleware stack changes are not applied
    def ht_setup_fn():
        try:
            instrumentation_wrapper.with_app(fastapi_app)
            agent_init.register_library(FAST_API_KEY, instrumentation_wrapper)
        except:  # pylint:disable=W0702
            logger.error('registering fastapi instrumentation in fastapi setup patch failed '
                         '- continuing without instrumenting fastapi')
        app = original_fn(fastapi_app)
        return app

    return ht_setup_fn()



"""
Agent configuration logic that pull in values from a defaults list,
environment variables, and the agent-config.yaml file.
"""
import logging
from google.protobuf import json_format as jf
from google.protobuf.wrappers_pb2 import BoolValue  # pylint:disable=E0611
from tyke.agent.config import config_pb2
from tyke.agent.config.default import *
from tyke.env_var_settings import get_env_value
from .file import load_config_from_file
from .environment import load_config_from_env


# Configuration attributes specific to pythonagent
PYTHON_SPECIFIC_ATTRIBUTES: list = [
    '_use_console_span_exporter'
]

# Initialize logger
logger = logging.getLogger(__name__)  # pylint: disable=C0103


def merge_config(base_config, overriding_config):
    """
    Returns the merged result of two configs recursively
    """

    for key in overriding_config:
        if key in base_config and isinstance(base_config[key], dict):
            if key in overriding_config:
                base_config[key] = merge_config(
                    base_config[key], overriding_config[key])
        else:
            if key == 'propagation_formats':
                logger.debug('Merging propagation_formats values.')
                if hasattr(base_config, key):
                    base_config[key] = set(
                        base_config[key].extend(overriding_config[key]))
                else:
                    base_config[key] = overriding_config[key]
            else:
                base_config[key] = overriding_config[key]
    return base_config

# Read agent-config file and override with environment variables as necessaary
class AgentConfig:  # pylint: disable=R0902,R0903
    '''A wrapper around the agent configuration logic'''

    def __init__(self):  # pylint: disable=R0912,R0915
        """
        Returns a new instance of config_pb2.AgentConfig when a new AgentConfig() is created.
        If 'TYKE_CONFIG_FILE' is specified in the environment data would be loaded from that file.
        If not, data would be loaded from 'DEFAULT_AGENT_CONFIG' on 'default.py'
        """

        config_dict = DEFAULT_AGENT_CONFIG
        custom_config = {}
        file_dict = _read_from_file()
        if file_dict is not None:
            config_dict = merge_config(config_dict, file_dict)

        env_dict = load_config_from_env()
        config_dict = merge_config(config_dict, env_dict)

        self.custom_config = _apply_custom_config_options(custom_config, config_dict)

        # transform(to not break old support) like zipkin != proto def of ZIPKIN
        _transform_values(config_dict)
        logger.info("Config init complete - config state: %s", config_dict)

        # Create protobuf Reporting object
        reporting = jf.Parse(jf.MessageToJson(
            config_pb2.Reporting()), config_pb2.Reporting)
        reporting.endpoint = config_dict['reporting']['endpoint']
        reporting.secure = config_dict['reporting']['secure']
        reporting.token = config_dict['reporting']['token']

        # Set trace_reporter_type
        if config_dict['reporting']['trace_reporter_type'] == 'OTLP':
            reporting.trace_reporter_type = config_pb2.TraceReporterType.OTLP
        elif config_dict['reporting']['trace_reporter_type'] == 'ZIPKIN':
            reporting.trace_reporter_type = config_pb2.TraceReporterType.ZIPKIN
        else:
            # Default to ZIPKIN
            reporting.trace_reporter_type = config_pb2.TraceReporterType.OTLP

        # Create DataCapture Message components
        rpc_body = config_pb2.Message(request=BoolValue(
            value=config_dict['data_capture']['rpc_body']['request']),
            response=BoolValue(
                value=config_dict['data_capture']['rpc_body']['response']))
        rpc_metadata = config_pb2.Message(request=BoolValue(
            value=config_dict['data_capture']['rpc_metadata']['request']),
            response=BoolValue(
                value=config_dict['data_capture']['rpc_metadata']['response']))
        http_body = config_pb2.Message(request=BoolValue(
            value=config_dict['data_capture']['http_body']['request']),
            response=BoolValue(
                value=config_dict['data_capture']['http_body']['response']))
        http_headers = config_pb2.Message(request=BoolValue(
            value=config_dict['data_capture']['http_headers']['request']),
            response=BoolValue(
                value=config_dict['data_capture']['http_headers']['response']))

        # Create Protobuf DataCapture object
        data_capture = jf.Parse(jf.MessageToJson(
            config_pb2.DataCapture()), config_pb2.DataCapture)
        data_capture.http_headers = http_headers
        data_capture.http_body = http_body
        data_capture.rpc_metadata = rpc_metadata
        data_capture.rpc_body = rpc_body
        data_capture.body_max_size_bytes = config_dict['data_capture']['body_max_size_bytes']

        # Create Protobuf AgentConfig object
        self.agent_config: config_pb2.AgentConfig = jf.Parse(jf.MessageToJson(
            config_pb2.AgentConfig()), config_pb2.AgentConfig)
        self.agent_config.service_name = config_dict['service_name']
        self.agent_config.reporting = reporting
        self.agent_config.data_capture = data_capture
        tmp_propagation_formats = []
        if 'TRACECONTEXT' in config_dict['propagation_formats']:
            tmp_propagation_formats.append(
                config_pb2.PropagationFormat.TRACECONTEXT)
            tmp_propagation_formats = list(set(tmp_propagation_formats))
        if 'B3' in config_dict['propagation_formats']:
            tmp_propagation_formats.append(config_pb2.PropagationFormat.B3)
            tmp_propagation_formats = list(set(tmp_propagation_formats))
        if not tmp_propagation_formats:
            # Default to TRACECONTEXT
            tmp_propagation_formats.append(
                config_pb2.PropagationFormat.TRACECONTEXT)
        self.agent_config.propagation_formats = tmp_propagation_formats
        self.agent_config.enabled = config_dict['enabled']

        self.agent_config.resource_attributes = config_dict['resource_attributes']

    def dump_config(self):
        '''Dump configuration information.'''
        logger.debug(self.__dict__)

    def use_console_span_exporter(self) -> bool:
        '''Initialize InMemorySpanExporter'''
        return self.custom_config.get('_use_console_span_exporter')


def _transform_values(config_dict):
    # This has to be upcase or else enum lookup fails
    config_dict['reporting']['trace_reporter_type'] = config_dict['reporting']['trace_reporter_type'].upper() # pylint:disable=C0301
    config_dict['propagation_formats'] = list(map(str.upper, config_dict['propagation_formats']))

def _apply_custom_config_options(current_custom, next_config):
    for key in PYTHON_SPECIFIC_ATTRIBUTES:
        if key in next_config:
            value = next_config.pop(key)
            current_custom[key] = value
    return current_custom

def _read_from_file():
    config_file = get_env_value('CONFIG_FILE')
    if config_file is None:
        logger.debug("no config file found")
        return None

    if len(config_file) == 0:  # pylint:disable=R1705:
        # TYKE_CONFIG_FILE can be passed as empty string which is invalid
        logger.error('Failed to load CONFIG_FILE env var is empty')
        return None
    else:
        config_from_file = load_config_from_file(config_file)
        logger.debug('Successfully loaded config file')
        return config_from_file



'''Tyke wrapper around OTel botocore Instrumentor''' # pylint: disable=R0801
import logging
from opentelemetry.instrumentation.botocore import BotocoreInstrumentor # pylint:disable=E0611,E0401
from tyke.agent.instrumentation import BaseInstrumentorWrapper


# Initialize logger with local module name
logger = logging.getLogger(__name__) # pylint: disable=C0103

class BotocoreInstrumentationWrapper(BotocoreInstrumentor, BaseInstrumentorWrapper):
    '''Tyke wrapper around OTel Botocore Instrumentor class'''



"""
Used to override env var prefix
these are separated out to prevent any classes from being imported before being customized
"""

import os

_DEFAULT_ENV_VAR_PREFIX = 'TYKE'
ENV_VAR_PREFIXES = [_DEFAULT_ENV_VAR_PREFIX]

def get_env_value(target_key):
    """checks env for given target_key with any configured prefix"""
    for prefix in ENV_VAR_PREFIXES:
        env_var_key = f"{prefix}_{target_key}"
        if env_var_key in os.environ:
            return os.environ[env_var_key]
    return None



'''Environment config loader'''

import logging

from tyke.env_var_settings import get_env_value

logger = logging.getLogger(__name__)  # pylint: disable=C0103


def load_config_from_env() -> dict:  # pylint: disable=R0912,R0915,R0914
    '''Loads config from environment variables'''
    config = {}

    service_name = get_env_value('SERVICE_NAME')
    if service_name:
        logger.debug("[env] Loaded SERVICE_NAME from env")
        config['service_name'] = service_name

    # Reporting
    config['reporting'] = {}

    reporting_endpoint = get_env_value('REPORTING_ENDPOINT')
    if reporting_endpoint:
        logger.debug("[env] Loaded REPORTING_ENDPOINT from env")
        config['reporting']['endpoint'] = reporting_endpoint

    reporter_type = get_env_value('REPORTING_TRACE_REPORTER_TYPE')
    if reporter_type:
        logger.debug(
            "[env] Loaded REPORTING_TRACE_REPORTER_TYPE from env")
        config['reporting']['trace_reporter_type'] = reporter_type

    reporting_secure = get_env_value('REPORTING_SECURE')
    if reporting_secure:
        logger.debug("[env] Loaded REPORTING_SECURE from env")
        config['reporting']['secure'] = reporting_secure.lower(
        ) == 'true'

    reporting_token = get_env_value('REPORTING_TOKEN')
    if reporting_token:
        logger.debug("[env] Loaded REPORTING_TOKEN from env")
        config['reporting']['token'] = reporting_token

    if len(config['reporting']) == 0:
        del config['reporting']

    # Data capture
    config['data_capture'] = {}

    config['data_capture']['http_headers'] = {}
    headers_request = get_env_value('DATA_CAPTURE_HTTP_HEADERS_REQUEST')
    if headers_request:
        logger.debug(
            "[env] Loaded DATA_CAPTURE_HTTP_HEADERS_REQUEST from env")
        config['data_capture']['http_headers']['request'] = _is_true(headers_request)

    headers_response = get_env_value('DATA_CAPTURE_HTTP_HEADERS_RESPONSE')
    if headers_response:
        logger.debug(
            "[env] Loaded DATA_CAPTURE_HTTP_HEADERS_RESPONSE from env")
        config['data_capture']['http_headers']['response'] = _is_true(headers_response)

    if len(config['data_capture']['http_headers']) == 0:
        del config['data_capture']['http_headers']

    config['data_capture']['http_body'] = {}
    body_request = get_env_value('DATA_CAPTURE_HTTP_BODY_REQUEST')
    if body_request:
        logger.debug(
            "[env] Loaded DATA_CAPTURE_HTTP_BODY_REQUEST from env")
        config['data_capture']['http_body']['request'] = _is_true(body_request)

    body_response = get_env_value('DATA_CAPTURE_HTTP_BODY_RESPONSE')
    if body_response:
        logger.debug(
            "[env] Loaded DATA_CAPTURE_HTTP_BODY_RESPONSE from env")
        config['data_capture']['http_body']['response'] = _is_true(body_response)

    if len(config['data_capture']['http_body']) == 0:
        del config['data_capture']['http_body']

    config['data_capture']['rpc_metadata'] = {}
    rpc_metadata_request = get_env_value('DATA_CAPTURE_RPC_METADATA_REQUEST')
    if rpc_metadata_request:
        logger.debug(
            "[env] Loaded DATA_CAPTURE_RPC_METADATA_REQUEST from env")
        config['data_capture']['rpc_metadata']['request'] = _is_true(rpc_metadata_request)

    rpc_metadata_response = get_env_value('DATA_CAPTURE_RPC_METADATA_RESPONSE')
    if rpc_metadata_response:
        logger.debug(
            "[env] Loaded DATA_CAPTURE_RPC_METADATA_RESPONSE from env")
        config['data_capture']['rpc_metadata']['response'] = _is_true(rpc_metadata_response)

    if len(config['data_capture']['rpc_metadata']) == 0:
        del config['data_capture']['rpc_metadata']

    config['data_capture']['rpc_body'] = {}
    rpc_body_request = get_env_value('DATA_CAPTURE_RPC_BODY_REQUEST')
    if rpc_body_request:
        logger.debug(
            "[env] Loaded DATA_CAPTURE_RPC_BODY_REQUEST from env")
        config['data_capture']['rpc_body']['request'] = _is_true(rpc_body_request)

    rpc_body_response = get_env_value('DATA_CAPTURE_RPC_BODY_RESPONSE')
    if rpc_body_response:
        logger.debug(
            "[env] Loaded DATA_CAPTURE_RPC_BODY_RESPONSE from env")
        config['data_capture']['rpc_body']['response'] = _is_true(rpc_body_response)

    if len(config['data_capture']['rpc_body']) == 0:
        del config['data_capture']['rpc_body']

    body_max_size_bytes = get_env_value('DATA_CAPTURE_BODY_MAX_SIZE_BYTES')
    if body_max_size_bytes:
        logger.debug(
            "[env] Loaded DATA_CAPTURE_BODY_MAX_SIZE_BYTES from env")
        config['data_capture']['body_max_size_bytes'] = int(body_max_size_bytes)

    if len(config['data_capture']) == 0:
        del config['data_capture']

    propagation_formats = get_env_value('PROPAGATION_FORMATS')
    if propagation_formats and len(propagation_formats) > 0:
        logger.debug("[env] Loaded PROPAGATION_FORMATS from env")
        config['propagation_formats'] = propagation_formats.split(',')

    enabled = get_env_value('ENABLED')
    if enabled:
        logger.debug("[env] Loaded ENABLED from env")
        config['enabled'] = _is_true(enabled)

    console_span_exporter = get_env_value('ENABLE_CONSOLE_SPAN_EXPORTER')
    if console_span_exporter:
        logger.debug("[env] Loaded ENABLE_CONSOLE_SPAN_EXPORTER from env")
        config['_use_console_span_exporter'] = _is_true(console_span_exporter)
    resource_attributes = get_env_value('RESOURCE_ATTRIBUTES')
    if resource_attributes:
        config['resource_attributes'] = {}
        logger.debug('[env] Loaded RESOURCE_ATTRIBUTES from env')
        groups = resource_attributes.split(',')
        for group in groups:
            key, value = group.split('=')
            config['resource_attributes'][key] = value
    return config

def _is_true(value):
    return value.lower() == 'true'



"""includes a middleware.__call__ wrapper and the fast api instrumentor implementation + hooks"""
import logging
from functools import wraps

from fastapi import HTTPException
from starlette.datastructures import Headers
from starlette.requests import Request
from starlette.responses import PlainTextResponse

from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor
from opentelemetry.instrumentation.asgi import OpenTelemetryMiddleware, set_status_code
from opentelemetry.instrumentation.asgi import get_host_port_url_tuple, \
    context, extract, asgi_getter, trace, collect_request_attributes

from tyke.agent.filter.registry import Registry, TYPE_HTTP
from tyke.agent.instrumentation import BaseInstrumentorWrapper

logger = logging.getLogger(__name__)  # pylint: disable=C0103


# We need to replace the entire __call__ method so that we can
# pull the req body from the original requests receive
async def replaced_ot_middleware_call(self, scope, receive, send): # pylint:disable=R0914
    """The ASGI application

    Args:
        scope: A ASGI environment.
        receive: An awaitable callable yielding dictionaries
        send: An awaitable callable taking a single dictionary as argument.
    """
    if scope["type"] not in ("http", "websocket"):
        return await self.app(scope, receive, send)

    _, _, url = get_host_port_url_tuple(scope)
    if self.excluded_urls and self.excluded_urls.url_disabled(url):
        return await self.app(scope, receive, send)

    token = context.attach(extract(scope, getter=asgi_getter))
    span_name, additional_attributes = self.default_span_details(scope)

    messages = []
    more_body = True
    while more_body:
        message = await receive()
        messages.append(message)
        more_body = message.get("more_body", False)
    body = b''.join([message.get("body", b"") for message in messages])

    try:
        with self.tracer.start_as_current_span(
                span_name,
                kind=trace.SpanKind.SERVER,
        ) as span:
            if span.is_recording():
                attributes = collect_request_attributes(scope)
                attributes.update(additional_attributes)
                for key, value in attributes.items():
                    span.set_attribute(key, value)

            should_forward_to_handler = True
            if callable(self.server_request_hook):
                should_forward_to_handler = self.server_request_hook(span, scope, body)

            @wraps(receive)
            async def wrapped_receive():
                with self.tracer.start_as_current_span(
                        " ".join((span_name, scope["type"], "receive"))
                ) as receive_span:
                    if messages:
                        return messages.pop(0)
                        # Once that's done we can just await any other messages.
                    if receive_span.is_recording():
                        if message["type"] == "websocket.receive":
                            set_status_code(receive_span, 200)
                        receive_span.set_attribute("type", message["type"])
                    return await receive()

            @wraps(send)
            async def wrapped_send(message):
                send_span = span
                if callable(self.client_response_hook):
                    self.client_response_hook(send_span, message)
                if send_span.is_recording():
                    if message["type"] == "http.response.start":
                        status_code = message["status"]
                        set_status_code(span, status_code)
                    elif message["type"] == "websocket.send":
                        set_status_code(span, 200)
                    elif message["type"] == 'http.response.body':
                        pass
                await send(message)

            if should_forward_to_handler:
                await self.app(scope, wrapped_receive, wrapped_send)
            else:
                set_status_code(span, 403)
                raise TykeException(status_code=403)
    finally:
        context.detach(token)


OpenTelemetryMiddleware.__call__ = replaced_ot_middleware_call


class TykeException(HTTPException):
    """An exception our middleware will raise if a filter determines to block the request"""

class FastAPIInstrumentorWrapper(BaseInstrumentorWrapper):
    """Used to instrument fast api apps once an app has been loaded """
    def __init__(self):
        super().__init__()
        self.app = None

    def with_app(self, app):
        """Need to defer instrumentation until we have a reference to the fast app
        during fastapi.FastAPI.setup this is invoked and stored so that we can then instrument"""
        self.app = app

    def instrument(self):
        """Adds our error handler as well tracing middleware"""
        if self.app is None:
            print("No app defined, cant instrument fast api")
            return

        async def catch_exceptions_middleware(request: Request, call_next):
            try:
                return await call_next(request)
            except TykeException:
                return PlainTextResponse(status_code=403)

        # configure fast api instrumentor w hooks
        FastAPIInstrumentor.instrument_app(app=self.app,
                                           server_request_hook=self.server_request_hook,
                                           client_response_hook=self.client_response_hook)

        # Need to add exception middleware to the top of middleware stack
        # if instrument_app is added after this the exceptions wont be handled by us but by default
        # fast api handler(which results in 500 instead of 403)
        self.app.middleware('http')(catch_exceptions_middleware)

    def server_request_hook(self, span, req_data, body):
        """this function is used to capture request attributes"""
        span.update_name(f"{req_data['method']} {span.name}")
        headers = dict(Headers(raw=req_data['headers']))
        request_url = str(Request(req_data).url)
        self.generic_request_handler(headers, body, span)

        block_result = Registry().apply_filters(span,
                                                request_url,
                                                headers,
                                                body,
                                                TYPE_HTTP)
        if block_result:
            logger.debug('should block evaluated to true, aborting with 403')
            return False
        return True

    def client_response_hook(self, span, resp_data):
        """used to capture the response data
        this function is called twice, once during each resp_phase"""
        resp_phase = resp_data['type']
        if resp_phase == "http.response.start":
            status_code = resp_data["status"]
            set_status_code(span, status_code)
            headers = dict(Headers(raw=resp_data['headers']))
            should_capture_body = self._capture_headers(self._process_response_headers,
                                                        self.HTTP_RESPONSE_HEADER_PREFIX,
                                                        span, headers, self._process_response_body)
            span.set_attribute('tyke.capture', should_capture_body)

        elif resp_phase == 'http.response.body':
            should_capture = span.attributes.get('tyke.capture')
            if should_capture:
                body_data = resp_data['body']
                body_str = None
                if isinstance(body_data, bytes):
                    body_str = body_data.decode('UTF8', 'backslashreplace')
                else:
                    body_str = body_data

                resp_body_str = self.grab_first_n_bytes(body_str)
                span.set_attribute('http.response.body', resp_body_str)

    def uninstrument(self):
        """Used to uninstrument fast api app"""
        if self.app:
            FastAPIInstrumentor.uninstrument_app(self.app)



'''Tyke wrapper around OTel Lambda Instrumentor'''  # pylint: disable=R0801
import logging
import os
from typing import Any, Callable
from wrapt import wrap_function_wrapper

from opentelemetry.instrumentation.aws_lambda import AwsLambdaInstrumentor
from opentelemetry.context.context import Context
from opentelemetry.instrumentation.aws_lambda.version import __version__
from opentelemetry.instrumentation.aws_lambda import _default_event_context_extractor
from opentelemetry.semconv.resource import ResourceAttributes
from opentelemetry.semconv.trace import SpanAttributes
from opentelemetry.trace import (
    SpanKind,
    TracerProvider,
    get_tracer,
    get_tracer_provider,
)
from opentelemetry.instrumentation import aws_lambda

from tyke.agent.instrumentation import BaseInstrumentorWrapper
logger = logging.getLogger(__name__)  # pylint: disable=C0103


class AwsLambdaInstrumentorWrapper(AwsLambdaInstrumentor, BaseInstrumentorWrapper):
    '''Tyke wrapper around OTel Lambda Instrumentor'''

    def __init__(self):
        '''Constructor'''
        logger.debug('Entering AwsLambdaInstrumentorWrapper.__init__().')
        super().__init__()

    def uninstrument(self, **kwargs):
        pass

    # We need to replace default _instrument behavior to capture request/resp data
    def _ht_instrument(self,  # pylint:disable=R0915
            wrapped_module_name,
            wrapped_function_name,
            event_context_extractor: Callable[[Any], Context],
            tracer_provider: TracerProvider = None,
    ):
        wrapper_instance = self
        def _instrumented_lambda_handler_call( # pylint:disable=R0914,R0915,R0912
                call_wrapped, _instance, args, kwargs
        ):

            orig_handler_name = ".".join(
                [wrapped_module_name, wrapped_function_name]
            )

            lambda_event = args[0]


            parent_context = aws_lambda._determine_parent_context(  # pylint:disable=W0212
                lambda_event, event_context_extractor
            )

            # See more:
            # https://docs.aws.amazon.com/lambda/latest/dg/with-sqs.html
            try:
                if lambda_event["Records"][0]["eventSource"] == "aws:sqs":
                    span_kind = SpanKind.CONSUMER
            except (IndexError, KeyError, TypeError):
                span_kind = SpanKind.SERVER

            tracer = get_tracer(__name__, __version__, tracer_provider)

            with tracer.start_as_current_span(
                    name=orig_handler_name,
                    context=parent_context,
                    kind=span_kind,
            ) as span:
                if span.is_recording():
                    logger.debug(lambda_event)

                    headers = lambda_event.get('headers', {})
                    multi_value_headers = lambda_event.get('multiValueHeaders', {})
                    for header_key in multi_value_headers:
                        headers[header_key] = ','.join(multi_value_headers[header_key])
                    if span.is_recording():
                        headers = lambda_event.get('headers', {})
                        lambda_request_context = lambda_event.get('requestContext', {})
                        logger.debug(lambda_request_context)

                    if lambda_request_context.get('http', None):
                        http_context = lambda_request_context.get('http')
                        span.set_attribute(SpanAttributes.HTTP_METHOD, http_context.get('method', None))
                        protocol = headers.get('x-forwarded-proto', None)
                        if protocol:
                            span.set_attribute(SpanAttributes.HTTP_SCHEME, protocol)
                        host = headers.get('host', None)
                        path = http_context.get('path', '')
                        span.set_attribute(SpanAttributes.HTTP_TARGET, path)
                        span.set_attribute(SpanAttributes.HTTP_HOST, host)
                        cookies = lambda_event.get('cookies', [])
                        if len(cookies) > 0:
                            cookie_header = ';'.join(cookies)
                            span.set_attribute('http.request.header.cookie', cookie_header)
                    elif lambda_request_context.get('path', None):
                        span.set_attribute(SpanAttributes.HTTP_METHOD, lambda_request_context.get('httpMethod', None))
                        span.set_attribute(SpanAttributes.HTTP_SCHEME, lambda_request_context.get('protocol', None))
                        protocol = headers.get('x-forwarded-proto', None)
                        if protocol:
                            span.set_attribute(SpanAttributes.HTTP_SCHEME, protocol)
                        host = headers.get('host', None) or \
                               lambda_event.get('multiValueHeaders', {}).get('Host', [None])[0]  # pylint:disable=C0301
                        path = lambda_request_context.get('path', '')
                        span.set_attribute(SpanAttributes.HTTP_TARGET, path)
                        span.set_attribute(SpanAttributes.HTTP_HOST, host)

                    wrapper_instance.generic_request_handler(headers, lambda_event.get('body', None), span)
                    lambda_context = args[1]
                    # NOTE: The specs mention an exception here, allowing the
                    # `ResourceAttributes.FAAS_ID` attribute to be set as a span
                    # attribute instead of a resource attribute.
                    #
                    # See more:
                    # https://github.com/open-telemetry/opentelemetry-specification/blob/main/specification/trace/semantic_conventions/faas.md#example
                    span.set_attribute(
                        ResourceAttributes.FAAS_ID,
                        lambda_context.invoked_function_arn,
                    )
                    span.set_attribute(
                        SpanAttributes.FAAS_EXECUTION,
                        lambda_context.aws_request_id,
                    )

                result = call_wrapped(*args, **kwargs)
                if result and isinstance(result, dict):
                    result_status_code = result.get('statusCode', None)

                    if result_status_code:
                        span.set_attribute(SpanAttributes.HTTP_STATUS_CODE, result_status_code)

                    wrapper_instance.generic_response_handler(result.get('headers', {}), result.get('body', None), span)
                else:
                    logger.info("Not processing lambda response, response is not a dict")
            _tracer_provider = tracer_provider or get_tracer_provider()
            try:
                # NOTE: `force_flush` before function quit in case of Lambda freeze.
                # Assumes we are using the OpenTelemetry SDK implementation of the
                # `TracerProvider`.
                logger.info('Attempting to export spans')
                _tracer_provider.force_flush(5000)  # max of 5 seconds to export
                logger.info('Exported spans')
            except Exception:  # pylint: disable=broad-except
                logger.error(
                    "TracerProvider was missing `force_flush` method. \
                    This is necessary in case of a Lambda freeze and would exist in the OTel SDK implementation."
                )

            return result
        logger.debug("Creating wrapped lambda_handler")
        wrap_function_wrapper(
            wrapped_module_name,
            wrapped_function_name,
            _instrumented_lambda_handler_call,
        )

    def _instrument(self, **kwargs):
        """Instruments Lambda Handlers on AWS Lambda.

        See more:
        https://github.com/open-telemetry/opentelemetry-specification/blob/main/specification/trace/semantic_conventions/instrumentation/aws-lambda.md#instrumenting-aws-lambda

        Args:
            **kwargs: Optional arguments
                ``tracer_provider``: a TracerProvider, defaults to global
                ``event_context_extractor``: a method which takes the Lambda
                    Event as input and extracts an OTel Context from it. By default,
                    the context is extracted from the HTTP headers of an API Gateway
                    request.
        """
        logger.debug("in AwsLambdaInstrumentorWrapper _instrument")
        lambda_handler = os.environ.get(aws_lambda.ORIG_HANDLER, os.environ.get(aws_lambda._HANDLER))  # pylint:disable=W0212
        if lambda_handler is None:
            logger.debug("Lambda handler function not defined.  This is expected in non-lambda environments")
            return
        # pylint: disable=attribute-defined-outside-init
        (
            self._wrapped_module_name,
            self._wrapped_function_name,
        ) = lambda_handler.rsplit(".", 1)

        self._ht_instrument(
            self._wrapped_module_name,
            self._wrapped_function_name,
            event_context_extractor=kwargs.get(
                "event_context_extractor", _default_event_context_extractor
            ),
            tracer_provider=kwargs.get("tracer_provider"),
        )



"""
Logging configuration
"""
import logging
import sys
import traceback

from tyke.env_var_settings import get_env_value

_LOG_LEVEL = {
    None: logging.INFO,
    '': logging.INFO,
    'INFO': logging.INFO,
    'DEBUG': logging.DEBUG,
    'ERROR': logging.ERROR,
    'WARNING': logging.WARNING,
    'CRTITICAL': logging.CRITICAL,
    'NOTSET': logging.NOTSET
}

def get_custom_logger(name: str) -> logging.Logger:
    '''Agent logger configuration'''
    try:
        formatter = logging.Formatter(fmt='%(asctime)s %(levelname)-8s %(message)s',
                                      datefmt='%Y-%m-%d %H:%M:%S')
        screen_handler = logging.StreamHandler(stream=sys.stdout)
        screen_handler.setFormatter(formatter)
        log_level = logging.INFO
        logger_ = logging.getLogger(name)
        env_value = get_env_value('LOG_LEVEL')
        if env_value:
            log_level = _LOG_LEVEL.get(env_value, log_level)

        logger_.setLevel(log_level)
        logger_.addHandler(screen_handler)
        return logger_
    except Exception as err:  # pylint: disable=W0703
        print('Failed to customize logger: exception=%s, stacktrace=%s',
              err,
              traceback.format_exc())
        return None



'''Base class of all Tyke Instrumentation Wrapper classes'''
import sys
import os.path
import inspect
import traceback
import json
import logging

from opentelemetry.attributes import BoundedAttributes
from opentelemetry.trace.span import Span

# Setup logger name
logger = logging.getLogger(__name__)  # pylint: disable=C0103


# This is a base class for all Tyke Instrumentation wrapper classes
class BaseInstrumentorWrapper:
    '''This is a base class for all Tyke Instrumentation wrapper classes'''
    # Standard extended span attribute names / prefixes
    HTTP_REQUEST_HEADER_PREFIX = 'http.request.header.'
    HTTP_RESPONSE_HEADER_PREFIX = 'http.response.header.'
    HTTP_REQUEST_BODY_PREFIX = 'http.request.body'
    HTTP_RESPONSE_BODY_PREFIX = 'http.response.body'
    RPC_REQUEST_METADATA_PREFIX = 'rpc.request.metadata.'
    RPC_RESPONSE_METADATA_PREFIX = 'rpc.response.metadata.'
    RPC_REQUEST_BODY_PREFIX = 'rpc.request.body'
    RPC_RESPONSE_BODY_PREFIX = 'rpc.response.body'

    # Constructor
    def __init__(self):
        '''constructor'''
        logger.debug('Entering BaseInstrumentorWrapper constructor.')
        super().__init__()
        self._process_request_headers = False
        self._process_response_headers = False
        self._process_request_body = False
        self._process_response_body = False
        self._max_body_size = 128 * 1024

    # Set whether request headers should be put in extended span, takes a BoolValue as input
    def set_process_request_headers(self, process_request_headers) -> None:
        '''Should it process request headers?'''
        logger.debug('Setting self._process_request_headers to \'%s\'',
                     process_request_headers.value)
        self._process_request_headers = process_request_headers

    # Set whether response headers should be put in extended span, takes a BoolValue as input
    def set_process_response_headers(self, process_response_headers) -> None:
        '''Should it process response headers?'''
        logger.debug('Setting self._process_response_headers to \'%s\'',
                     process_response_headers.value)
        self._process_response_headers = process_response_headers

    # Set whether request body should be put in extended span, takes a BoolValue as input
    def set_process_request_body(self, process_request_body) -> None:
        '''should it process request body?'''
        logger.debug('Setting self._process_request_body to \'%s\'',
                     process_request_body.value)
        self._process_request_body = process_request_body

    # Set whether response body should be put in extended span, takes a BoolValue as input
    def set_process_response_body(self, process_response_body) -> None:
        '''should it process response body?'''
        logger.debug('Setting self._process_response_body to \'%s\'',
                     process_response_body.value)
        self._process_response_body = process_response_body

    # Set max body size
    def set_body_max_size(self, max_body_size) -> None:
        '''Set the max body size for this instrumentor.'''
        logger.debug('Setting self.body_max_size to %s.', max_body_size)
        self._max_body_size = max_body_size

    # we need the headers lowercased multiple times
    # just do it once upfront
    def lowercase_headers(self, headers): # pylint:disable=R0201
        '''convert all headers to lowercase'''
        return {k.lower(): v for k, v in headers.items()}

    def add_headers_to_span(self, prefix: str, span: Span, headers: dict): # pylint:disable=R0201
        '''set header attributes on the span'''
        for header_key, header_value in headers.items():
            span.set_attribute(f"{prefix}{header_key}", header_value)

    _ALLOWED_CONTENT_TYPES = [
        "application/json",
        "application/graphql",
        'application/x-www-form-urlencoded'
    ]

    # We need the content type to do some escaping
    # so if we return a content type, that indicates valid for capture,
    # otherwise don't capture

    # Bhaskar modified
    def eligible_based_on_content_type(self, headers: dict):
        """find content-type in headers"""
        content_type = headers.get("content-type")
        if type(content_type) is tuple:
            content_type = content_type[1]
        return (
            content_type if content_type in self._ALLOWED_CONTENT_TYPES else None
        )  # plyint:disable=R1710

    # def eligible_based_on_content_type(self, headers: dict):
    #     '''find content-type in headers'''
    #     content_type = headers.get("content-type")
    #     return content_type in self._ALLOWED_CONTENT_TYPES

    def _capture_headers(self, record_headers: bool, header_prefix: str, # pylint:disable=R0913
                         span, headers: dict, record_body):
        try:  # pylint: disable=R1702
            if not span.is_recording():
                return False

            logger.debug('Span is Recording!')
            lowercased_headers = self.lowercase_headers(headers)
            if record_headers:
                self.add_headers_to_span(header_prefix, span, lowercased_headers)

            if record_body:
                content_type_recordable = self.eligible_based_on_content_type(lowercased_headers)
                return content_type_recordable
            return False
        except:  # pylint: disable=W0702
            logger.debug('An error occurred in _capture_headers: exception=%s, stacktrace=%s',
                         sys.exc_info()[0],
                         traceback.format_exc())
            return False


    def _generic_handler(self, record_headers: bool, header_prefix: str, # pylint:disable=R0913
                         record_body: bool, body_prefix: str,
                         span: Span, headers: dict, body):
        logger.debug('Entering BaseInstrumentationWrapper.generic_handler().')
        try:  # pylint: disable=R1702
            if not span.is_recording():
                return span

            logger.debug('Span is Recording!')
            lowercased_headers = self.lowercase_headers(headers)
            if record_headers:
                self.add_headers_to_span(header_prefix, span, lowercased_headers)

            if record_body:
                content_type = self.eligible_based_on_content_type(lowercased_headers)

                if content_type is False:
                    return span

                body_str = None
                if isinstance(body, bytes):
                    body_str = body.decode('UTF8', 'backslashreplace')
                else:
                    body_str = body

                request_body_str = self.grab_first_n_bytes(body_str)
                span.set_attribute(body_prefix, request_body_str)

        except:  # pylint: disable=W0702
            logger.debug('An error occurred in genericRequestHandler: exception=%s, stacktrace=%s',
                         sys.exc_info()[0],
                         traceback.format_exc())
        finally:
            return span  # pylint: disable=W0150

    # Generic HTTP Request Handler
    def generic_request_handler(self,  # pylint: disable=R0912
                                request_headers: dict,
                                request_body,
                                span: Span) -> Span:
        '''Add extended request data to the span'''
        logger.debug('Entering BaseInstrumentationWrapper.genericRequestHandler().')
        return self._generic_handler(self._process_request_headers, self.HTTP_REQUEST_HEADER_PREFIX,
                                     self._process_request_body, self.HTTP_REQUEST_BODY_PREFIX,
                                     span, request_headers, request_body)

    # Generic HTTP Response Handler
    def generic_response_handler(self,  # pylint: disable=R0912
                                 response_headers: dict,
                                 response_body,
                                 span: Span) -> Span:  # pylint: disable=R0912
        '''generic response handler'''
        logger.debug(
            'Entering BaseInstrumentationWrapper.genericResponseHandler().')
        return self._generic_handler(self._process_response_headers, self.HTTP_RESPONSE_HEADER_PREFIX,
                                     self._process_response_body, self.HTTP_RESPONSE_BODY_PREFIX,
                                     span, response_headers, response_body)

    # Generic RPC Request Handler
    def generic_rpc_request_handler(self,
                                    request_headers: dict,
                                    request_body,
                                    span: Span) -> Span:
        '''Add extended request rpc data to span.'''
        logger.debug(
            'Entering BaseInstrumentationWrapper.genericRpcRequestHandler().')
        try:
            # Is the span currently recording?
            if not span.is_recording():
                return span

            logger.debug('Span is Recording!')
            lowercased_headers = self.lowercase_headers(request_headers)

            # Log rpc metatdata if requested
            if self._process_request_headers:
                self.add_headers_to_span(self.RPC_REQUEST_METADATA_PREFIX, span, lowercased_headers)
            # Log rpc body if requested
            if self._process_response_body:
                request_body_str = str(request_body)
                request_body_str = self.grab_first_n_bytes(request_body_str)
                span.set_attribute(self.RPC_REQUEST_BODY_PREFIX,
                                   request_body_str)
        except:  # pylint: disable=W0702
            logger.debug('An error occurred in genericRequestHandler: exception=%s, stacktrace=%s',
                         sys.exc_info()[0],
                         traceback.format_exc())
            # Not rethrowing to avoid causing runtime errors
        finally:
            return span  # pylint: disable=W0150

    # Generic RPC Response Handler
    def generic_rpc_response_handler(self,
                                     response_headers: dict,
                                     response_body,
                                     span: Span) -> Span:
        '''Add extended response rpc data to span'''
        logger.debug(
            'Entering BaseInstrumentationWrapper.genericRpcResponseHandler().')
        try:
            # is the span currently recording?
            if not span.is_recording():
                return span

            logger.debug('Span is Recording!')
            lowercased_headers = self.lowercase_headers(response_headers)
            # Log rpc metadata if requested?
            if self._process_response_headers:

                logger.debug('Dumping Response Headers:')
                self.add_headers_to_span(self.RPC_RESPONSE_METADATA_PREFIX, span, lowercased_headers)
            # Log rpc body if requested
            if self._process_response_body:
                response_body_str = str(response_body)
                logger.debug('Processing response body')
                response_body_str = self.grab_first_n_bytes(response_body_str)
                span.set_attribute(
                    self.RPC_RESPONSE_BODY_PREFIX, response_body_str)
        except:  # pylint: disable=W0702
            logger.debug('An error occurred in genericResponseHandler: exception=%s, stacktrace=%s',
                         sys.exc_info()[0],
                         traceback.format_exc())
            # Not rethrowing to avoid causing runtime errors
        finally:
            return span  # pylint: disable=W0150

    # Check body size
    def check_body_size(self, body: str) -> bool:
        '''Is the size of this message body larger than the configured max?'''
        if body in (None, ''):
            return False
        body_len = len(body)
        max_body_size = self._max_body_size
        if max_body_size and body_len > max_body_size:
            logger.debug('message body size is greater than max size.')
            return True
        return False

    # grab first N bytes
    def grab_first_n_bytes(self, body: str) -> str:
        '''Return the first N (max_body_size) bytes of a request'''
        if body in (None, ''):
            return ''
        if self.check_body_size(body):  # pylint: disable=R1705
            return body[0, self._max_body_size]
        else:
            return body



'''Tyke django instrumentor module wrapper.''' # pylint: disable=R0401
from email import header
import logging
import traceback

from opentelemetry.instrumentation.django import DjangoInstrumentor
from opentelemetry.trace import Span
from django.core.exceptions import PermissionDenied  # pylint:disable=C0415

from tyke.agent import constants
from tyke.agent.filter.registry import Registry, TYPE_HTTP
from tyke.agent.instrumentation import BaseInstrumentorWrapper


logger = logging.getLogger(__name__)  # pylint: disable=C0103


class DjangoInstrumentationWrapper(BaseInstrumentorWrapper):
    """wrapped class around django instrumentation"""
    def instrument(self):
        """configure django instrumentor w hooks"""
        DjangoInstrumentor().instrument(request_hook=self.request_hook,
                                        response_hook=self.response_hook)

    def uninstrument(self): # pylint:disable=R0201
        """need this to match wrapper interface for specs"""
        return

    def request_hook(self, span: Span, request):
        """django request hook before request is processed by app"""
        try:
            body = request.body
            headers = None
            if hasattr(request, "headers"):
                headers = request.headers
            elif hasattr(request, "_headers"):
                headers = request._headers
            span.update_name(f"{request.method} {span.name}")
            self.generic_request_handler(headers, body, span)
            full_url = request.build_absolute_uri()
            block_result = Registry().apply_filters(span,
                                                    full_url,
                                                    headers,
                                                    body,
                                                    TYPE_HTTP)
            if block_result:
                logger.debug('should block evaluated to true, aborting with 403')
                # since middleware chain is halted the status code is not set when blocked
                span.set_attribute('http.status_code', 403)
                span.end()
                raise PermissionDenied
        except PermissionDenied as permission_denied:
            raise permission_denied
        except Exception as err:  # pylint:disable=W0703
            logger.debug(constants.INST_RUNTIME_EXCEPTION_MSSG,
                         'django request hook',
                         err,
                         traceback.format_exc())


    def response_hook(self, span, _request, response):
        """django response hook before response is written out"""
        try:
            body = response.content
            headers = None
            if hasattr(response, "headers"):
                headers = response.headers
            elif hasattr(response, "_headers"):
                headers = response._headers
            self.generic_response_handler(headers, body, span)
        except Exception as err:  # pylint:disable=W0703
            logger.debug(constants.INST_RUNTIME_EXCEPTION_MSSG,
                         'django response hook',
                         err,
                         traceback.format_exc())



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: config.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0c\x63onfig.proto\x12\x1atykevision.agent.config.v1\x1a\x1egoogle/protobuf/wrappers.proto\"\x86\x04\n\x0b\x41gentConfig\x12\x32\n\x0cservice_name\x18\x01 \x01(\x0b\x32\x1c.google.protobuf.StringValue\x12\x38\n\treporting\x18\x02 \x01(\x0b\x32%.tykevision.agent.config.v1.Reporting\x12=\n\x0c\x64\x61ta_capture\x18\x03 \x01(\x0b\x32\'.tykevision.agent.config.v1.DataCapture\x12J\n\x13propagation_formats\x18\x04 \x03(\x0e\x32-.tykevision.agent.config.v1.PropagationFormat\x12+\n\x07\x65nabled\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.BoolValue\x12\x38\n\tjavaagent\x18\x06 \x01(\x0b\x32%.tykevision.agent.config.v1.JavaAgent\x12\\\n\x13resource_attributes\x18\x07 \x03(\x0b\x32?.tykevision.agent.config.v1.AgentConfig.ResourceAttributesEntry\x1a\x39\n\x17ResourceAttributesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x96\x03\n\tReporting\x12.\n\x08\x65ndpoint\x18\x01 \x01(\x0b\x32\x1c.google.protobuf.StringValue\x12*\n\x06secure\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.BoolValue\x12+\n\x05token\x18\x03 \x01(\x0b\x32\x1c.google.protobuf.StringValue\x12J\n\x13trace_reporter_type\x18\x05 \x01(\x0e\x32-.tykevision.agent.config.v1.TraceReporterType\x12/\n\tcert_file\x18\x06 \x01(\x0b\x32\x1c.google.protobuf.StringValue\x12\x35\n\x0fmetric_endpoint\x18\x07 \x01(\x0b\x32\x1c.google.protobuf.StringValue\x12L\n\x14metric_reporter_type\x18\x08 \x01(\x0e\x32..tykevision.agent.config.v1.MetricReporterType\"d\n\x07Message\x12+\n\x07request\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.BoolValue\x12,\n\x08response\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.BoolValue\"\xf1\x02\n\x0b\x44\x61taCapture\x12\x39\n\x0chttp_headers\x18\x01 \x01(\x0b\x32#.tykevision.agent.config.v1.Message\x12\x36\n\thttp_body\x18\x02 \x01(\x0b\x32#.tykevision.agent.config.v1.Message\x12\x39\n\x0crpc_metadata\x18\x03 \x01(\x0b\x32#.tykevision.agent.config.v1.Message\x12\x35\n\x08rpc_body\x18\x04 \x01(\x0b\x32#.tykevision.agent.config.v1.Message\x12\x38\n\x13\x62ody_max_size_bytes\x18\x05 \x01(\x0b\x32\x1b.google.protobuf.Int32Value\x12\x43\n\x1e\x62ody_max_processing_size_bytes\x18\x06 \x01(\x0b\x32\x1b.google.protobuf.Int32Value\"C\n\tJavaAgent\x12\x36\n\x10\x66ilter_jar_paths\x18\x01 \x03(\x0b\x32\x1c.google.protobuf.StringValue*-\n\x11PropagationFormat\x12\x06\n\x02\x42\x33\x10\x00\x12\x10\n\x0cTRACECONTEXT\x10\x01*`\n\x11TraceReporterType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06ZIPKIN\x10\x01\x12\x08\n\x04OTLP\x10\x02\x12\x0b\n\x07LOGGING\x10\x03\x12\x08\n\x04NONE\x10\x04\x12\r\n\tOTLP_HTTP\x10\x05*\xbf\x01\n\x12MetricReporterType\x12$\n METRIC_REPORTER_TYPE_UNSPECIFIED\x10\x00\x12\x1d\n\x19METRIC_REPORTER_TYPE_OTLP\x10\x01\x12#\n\x1fMETRIC_REPORTER_TYPE_PROMETHEUS\x10\x02\x12 \n\x1cMETRIC_REPORTER_TYPE_LOGGING\x10\x03\x12\x1d\n\x19METRIC_REPORTER_TYPE_NONE\x10\x04\x42 \n\x1eorg.tykevision.agent.config.v1b\x06proto3')

_PROPAGATIONFORMAT = DESCRIPTOR.enum_types_by_name['PropagationFormat']
PropagationFormat = enum_type_wrapper.EnumTypeWrapper(_PROPAGATIONFORMAT)
_TRACEREPORTERTYPE = DESCRIPTOR.enum_types_by_name['TraceReporterType']
TraceReporterType = enum_type_wrapper.EnumTypeWrapper(_TRACEREPORTERTYPE)
_METRICREPORTERTYPE = DESCRIPTOR.enum_types_by_name['MetricReporterType']
MetricReporterType = enum_type_wrapper.EnumTypeWrapper(_METRICREPORTERTYPE)
B3 = 0
TRACECONTEXT = 1
UNSPECIFIED = 0
ZIPKIN = 1
OTLP = 2
LOGGING = 3
NONE = 4
OTLP_HTTP = 5
METRIC_REPORTER_TYPE_UNSPECIFIED = 0
METRIC_REPORTER_TYPE_OTLP = 1
METRIC_REPORTER_TYPE_PROMETHEUS = 2
METRIC_REPORTER_TYPE_LOGGING = 3
METRIC_REPORTER_TYPE_NONE = 4


_AGENTCONFIG = DESCRIPTOR.message_types_by_name['AgentConfig']
_AGENTCONFIG_RESOURCEATTRIBUTESENTRY = _AGENTCONFIG.nested_types_by_name['ResourceAttributesEntry']
_REPORTING = DESCRIPTOR.message_types_by_name['Reporting']
_MESSAGE = DESCRIPTOR.message_types_by_name['Message']
_DATACAPTURE = DESCRIPTOR.message_types_by_name['DataCapture']
_JAVAAGENT = DESCRIPTOR.message_types_by_name['JavaAgent']
AgentConfig = _reflection.GeneratedProtocolMessageType('AgentConfig', (_message.Message,), {

  'ResourceAttributesEntry' : _reflection.GeneratedProtocolMessageType('ResourceAttributesEntry', (_message.Message,), {
    'DESCRIPTOR' : _AGENTCONFIG_RESOURCEATTRIBUTESENTRY,
    '__module__' : 'config_pb2'
    # @@protoc_insertion_point(class_scope:tykevision.agent.config.v1.AgentConfig.ResourceAttributesEntry)
    })
  ,
  'DESCRIPTOR' : _AGENTCONFIG,
  '__module__' : 'config_pb2'
  # @@protoc_insertion_point(class_scope:tykevision.agent.config.v1.AgentConfig)
  })
_sym_db.RegisterMessage(AgentConfig)
_sym_db.RegisterMessage(AgentConfig.ResourceAttributesEntry)

Reporting = _reflection.GeneratedProtocolMessageType('Reporting', (_message.Message,), {
  'DESCRIPTOR' : _REPORTING,
  '__module__' : 'config_pb2'
  # @@protoc_insertion_point(class_scope:tykevision.agent.config.v1.Reporting)
  })
_sym_db.RegisterMessage(Reporting)

Message = _reflection.GeneratedProtocolMessageType('Message', (_message.Message,), {
  'DESCRIPTOR' : _MESSAGE,
  '__module__' : 'config_pb2'
  # @@protoc_insertion_point(class_scope:tykevision.agent.config.v1.Message)
  })
_sym_db.RegisterMessage(Message)

DataCapture = _reflection.GeneratedProtocolMessageType('DataCapture', (_message.Message,), {
  'DESCRIPTOR' : _DATACAPTURE,
  '__module__' : 'config_pb2'
  # @@protoc_insertion_point(class_scope:tykevision.agent.config.v1.DataCapture)
  })
_sym_db.RegisterMessage(DataCapture)

JavaAgent = _reflection.GeneratedProtocolMessageType('JavaAgent', (_message.Message,), {
  'DESCRIPTOR' : _JAVAAGENT,
  '__module__' : 'config_pb2'
  # @@protoc_insertion_point(class_scope:tykevision.agent.config.v1.JavaAgent)
  })
_sym_db.RegisterMessage(JavaAgent)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\036org.tykevision.agent.config.v1'
  _AGENTCONFIG_RESOURCEATTRIBUTESENTRY._options = None
  _AGENTCONFIG_RESOURCEATTRIBUTESENTRY._serialized_options = b'8\001'
  _PROPAGATIONFORMAT._serialized_start=1549
  _PROPAGATIONFORMAT._serialized_end=1594
  _TRACEREPORTERTYPE._serialized_start=1596
  _TRACEREPORTERTYPE._serialized_end=1692
  _METRICREPORTERTYPE._serialized_start=1695
  _METRICREPORTERTYPE._serialized_end=1886
  _AGENTCONFIG._serialized_start=77
  _AGENTCONFIG._serialized_end=595
  _AGENTCONFIG_RESOURCEATTRIBUTESENTRY._serialized_start=538
  _AGENTCONFIG_RESOURCEATTRIBUTESENTRY._serialized_end=595
  _REPORTING._serialized_start=598
  _REPORTING._serialized_end=1004
  _MESSAGE._serialized_start=1006
  _MESSAGE._serialized_end=1106
  _DATACAPTURE._serialized_start=1109
  _DATACAPTURE._serialized_end=1478
  _JAVAAGENT._serialized_start=1480
  _JAVAAGENT._serialized_end=1547
# @@protoc_insertion_point(module_scope)



"""Filter abstract base class for blocking"""
from abc import ABC, abstractmethod
from opentelemetry.trace.span import Span


class Filter(ABC):
    """
    Filter evaluates whether request should be blocked, True blocks the request and False
    continues it.
    """
    @abstractmethod
    def evaluate_url_and_headers(self, span: Span, url: str, headers: dict, request_type) -> bool:
        """evaluate_url_and_headers can be used to evaluate both URL and Header"""

    @abstractmethod
    def evaluate_body(self, span: Span, body, headers: dict, request_type) -> bool:
        """evaluate_body can be used to evaluate the body content"""


class NoopFilter(Filter):
    """NoopFilter is a filter that never blocks"""

    def evaluate_url_and_headers(self, span: Span, url: str, headers: dict, request_type) -> bool:
        return False

    def evaluate_body(self, span: Span, body, headers: dict, request_type) -> bool:
        return False



'''Tyke wrapper around OTel postgresql instrumentor'''
import sys
import os.path
import logging
import traceback
from opentelemetry.instrumentation.psycopg2 import Psycopg2Instrumentor
from tyke.agent.instrumentation import BaseInstrumentorWrapper

# Initialize logger with local module name
logger = logging.getLogger(__name__) # pylint: disable=C0103

class PostgreSQLInstrumentorWrapper(Psycopg2Instrumentor, BaseInstrumentorWrapper):
    '''Tyke wrapper around OTel postgresql instrumentor class'''



'''Main entry point for Tyke agent module.'''
import os
import sys
import threading
import traceback
from contextlib import contextmanager

from deprecated import deprecated
import opentelemetry.trace as ot

from tyke.agent.instrumentation.instrumentation_definitions import SUPPORTED_LIBRARIES, \
    get_instrumentation_wrapper, REQUESTS_KEY, GRPC_CLIENT_KEY, DJANGO_KEY, MYSQL_KEY, GRPC_SERVER_KEY, \
    POSTGRESQL_KEY, AIOHTTP_CLIENT_KEY, FLASK_KEY, LAMBDA,FAST_API_KEY, _uninstrument_all
from tyke.env_var_settings import get_env_value
from tyke.agent.init import AgentInit
from tyke.agent.config import AgentConfig
from tyke.agent import constants
from tyke.agent import custom_logger
from tyke.version import __version__
# The Tyke Python Agent class

logger = custom_logger.get_custom_logger(__name__)


class Agent:
    '''Top-level entry point for Tyke agent.'''
    _instance = None
    _singleton_lock = threading.Lock()

    def __new__(cls):
        '''constructor'''
        if cls._instance is None:
            with cls._singleton_lock:
                logger.debug('Creating Agent')
                logger.debug('Python version: %s', sys.version)
                logger.debug('Tyke Agent version: %s', __version__)
                cls._instance = super(Agent, cls).__new__(cls)
                cls._instance._initialized = False
        else:
            logger.debug('Using existing Agent.')
        return cls._instance

    def __init__(self):
        '''Initializer'''
        self._initialized = False
        if not self._initialized:  # pylint: disable=E0203:
            logger.debug('Initializing Agent.')
            if not self.is_enabled():
                return
            try:
                self._config = AgentConfig()
                self._init = AgentInit(self._config)
                self._initialized = True
            except Exception as err:  # pylint: disable=W0703
                logger.error('Failed to initialize Agent: exception=%s, stacktrace=%s',
                             err,
                             traceback.format_exc())

    @contextmanager
    def edit_config(self):
        """Used by end users to modify the config"""
        try:
            # need to explicitly set this as None when modifying the config via code
            # to regenerate Trace Provider with new options
            ot._TRACER_PROVIDER = None  # pylint:disable=W0212
            agent_config = self._config.agent_config
            yield agent_config
            self._config.agent_config = agent_config
        finally:
            self._init.apply_config(self._config)

    def instrument(self, app=None, skip_libraries=None, auto_instrument=False):
        '''used to register applicable instrumentation wrappers'''
        if skip_libraries is None:
            skip_libraries = []
        if not self.is_initialized():
            logger.debug('agent is not initialized, not instrumenting')
            return

        for library_key in SUPPORTED_LIBRARIES:
            if library_key in skip_libraries:
                logger.debug('not attempting to instrument %s', library_key)
                continue

            self._instrument(library_key, app, auto_instrument)

    def _instrument(self, library_key, app=None, auto_instrument=False):
        """only used to allow the deprecated register_x library methods to still work"""
        wrapper_instance = get_instrumentation_wrapper(library_key)
        if wrapper_instance is None:
            return

        # Flask is a special case compared to rest of instrumentation
        # using tyke-instrument we can replace flask class def before app is initialized
        # however during code based instr we wrap the existing app
        # since replacing class def after app is initialized doesnt have an effect
        # the user has to pass the app in to agent.instrument()
        # we could resolve this edge case by instead having users directly add the middleware
        # ex: app = Flask();
        # app = TykeMiddleware(App) => this in turn does agent.instrument()
        # + we have ref to app
        if library_key == FLASK_KEY and app is not None:
            wrapper_instance.with_app(app)

        # since ht sitecustomize pushes the agent to the front of the load path django instrumentation will error
        # when using autoinstrumentation since we need the django app settings loaded to add middleware
        #
        # in order to resolve this if we detect django we wrap the wsgi/asgi app getter
        # and instrument as soon as the app is retrieved(since settings have to be configured
        # before returning loaded app)
        if library_key == DJANGO_KEY and auto_instrument is True:
            from tyke.agent.instrumentation.django.django_auto_instrumentation_compat import \
                add_django_auto_instr_wrappers  # pylint: disable=C0415
            add_django_auto_instr_wrappers(self, wrapper_instance)
            return

        # For FastAPI we need a handle to the user app before we can instrument fast & inject middleware
        # to make things easier for the user always add the below instrumentation wrappers
        # when FastAPI calls `.setup` take the app and then add instrumentation
        if library_key == FAST_API_KEY:
            from tyke.agent.instrumentation.fast_api.fast_api_auto_instrumentation_compat import \
                add_fast_api_auto_instr_wrappers # pylint: disable=C0415
            add_fast_api_auto_instr_wrappers(self, wrapper_instance)
            return

        if library_key == LAMBDA and '_HANDLER' not in os.environ:
            return

        self.register_library(library_key, wrapper_instance)

    def register_library(self, library_name, wrapper_instance):
        """will configure settings on an instrumentation wrapper + apply"""
        logger.debug('attempting to register library instrumentation: %s', library_name)
        try:
            self._init.init_library_instrumentation(library_name, wrapper_instance)
        except Exception as err:  # pylint: disable=W0703
            logger.debug(constants.EXCEPTION_MESSAGE, library_name, err, traceback.format_exc())

    def register_processor(self, processor) -> None:  # pylint: disable=R1710
        '''Add additional span exporters + processors'''
        logger.debug('Entering Agent.register_processor().')
        logger.debug("initialized %s", self.is_initialized())
        if not self.is_initialized():
            return None
        return self._init.register_processor(processor)

    def is_enabled(self) -> bool:  # pylint: disable=R0201
        '''Is agent enabled?'''
        enabled = get_env_value('ENABLED')
        if enabled:
            if enabled.lower() == 'false':
                logger.debug("ENABLED is disabled.")
                return False
        return True

    def is_initialized(self) -> bool:  # pylint: disable=R0201
        '''Is agent initialized - if an agent fails to init we should let the app continue'''
        if not self.is_enabled():
            return False
        if not self._initialized:
            return False
        return True

    # These methods are deprecated and replaced by a single call to `agent_instance.instrument()`
    AGENT_INSTRUMENT_INSTEAD = "You should use agent.instrument() instead"
    AGENT_INSTRUMENT_VERSION = '0.1.0'

    @deprecated(version=AGENT_INSTRUMENT_VERSION, reason=AGENT_INSTRUMENT_INSTEAD)
    def register_requests(self):
        """just a proxy to support deprecated method for instrumenting requests"""
        self._instrument(REQUESTS_KEY)

    @deprecated(version=AGENT_INSTRUMENT_VERSION, reason=AGENT_INSTRUMENT_INSTEAD)
    def register_grpc_client(self):
        """just a proxy to support deprecated method for instrumenting grpc_client"""
        self._instrument(GRPC_CLIENT_KEY)

    @deprecated(version=AGENT_INSTRUMENT_VERSION, reason=AGENT_INSTRUMENT_INSTEAD)
    def register_django(self):
        """just a proxy to support deprecated method for instrumenting django"""
        self._instrument(DJANGO_KEY)

    @deprecated(version=AGENT_INSTRUMENT_VERSION, reason=AGENT_INSTRUMENT_INSTEAD)
    def register_mysql(self):
        """just a proxy to support deprecated method for instrumenting mysql"""
        self._instrument(MYSQL_KEY)

    @deprecated(version=AGENT_INSTRUMENT_VERSION, reason=AGENT_INSTRUMENT_INSTEAD)
    def register_grpc_server(self):
        """just a proxy to support deprecated method for instrumenting grpc server"""
        self._instrument(GRPC_SERVER_KEY)

    @deprecated(version=AGENT_INSTRUMENT_VERSION, reason=AGENT_INSTRUMENT_INSTEAD)
    def register_postgresql(self):
        """just a proxy to support deprecated method for instrumenting postgresql"""
        self._instrument(POSTGRESQL_KEY)

    @deprecated(version=AGENT_INSTRUMENT_VERSION, reason=AGENT_INSTRUMENT_INSTEAD)
    def register_aiohttp_client(self):
        """just a proxy to support deprecated method for instrumenting aiohttp"""
        self._instrument(AIOHTTP_CLIENT_KEY)

    @deprecated(version=AGENT_INSTRUMENT_VERSION, reason=AGENT_INSTRUMENT_INSTEAD)
    def register_flask_app(self, app=None):
        """just a proxy to support deprecated method for instrumenting flask"""
        self._instrument(FLASK_KEY, app)



'''Tyke wrapper around OTel boto Instrumentor''' # pylint: disable=R0801
import logging
from opentelemetry.instrumentation.boto import BotoInstrumentor  # pylint:disable=E0611,E0401
from tyke.agent.instrumentation import BaseInstrumentorWrapper


# Initialize logger with local module name
logger = logging.getLogger(__name__) # pylint: disable=C0103

class BotoInstrumentationWrapper(BotoInstrumentor, BaseInstrumentorWrapper):
    '''Tyke wrapper around OTel Boto Instrumentor class'''



import sys

from setuptools import setup, find_packages

version_tuple = sys.version_info
# install_requires doesn't always work(like if on older versions of pip)
if sys.version_info < (3, 6):
    print("Tyke is not supported on python versions before 3.6")
    sys.exit(1)

exec(open('src/tyke/version.py').read())

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="tyke-agent",
    version=__version__,
    author="TykeVision",
    author_email="tech@tyke.ai",
    description="Tyke Python Agent",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tykevision/python-agent",
    project_urls={
        "Bug Tracker": "https://github.com/tykevision/python-agent/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent"
    ],
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=">=3.6",
    install_requires=[
        "opentelemetry-api==1.8.0",
        "opentelemetry-exporter-otlp==1.8.0",
        "opentelemetry-exporter-zipkin==1.8.0",
        "opentelemetry-instrumentation==0.27b0",
        "opentelemetry-instrumentation-aiohttp-client==0.27b0",
        "opentelemetry-instrumentation-boto==0.27b0",
        "opentelemetry-instrumentation-botocore==0.27b0",
        "opentelemetry-instrumentation-wsgi==0.27b0",
        "opentelemetry-instrumentation-fastapi==0.27b0",
        "opentelemetry-instrumentation-flask==0.27b0",
        "opentelemetry-instrumentation-mysql==0.27b0",
        "opentelemetry-instrumentation-psycopg2==0.27b0",
        "opentelemetry-instrumentation-requests==0.27b0",
        "opentelemetry-instrumentation-grpc==0.27b0",
        "opentelemetry-instrumentation-django==0.27b0",
        "opentelemetry-instrumentation-aws-lambda==0.27b0",
        "opentelemetry-propagator-b3==1.8.0",
        "opentelemetry-sdk==1.8.0",
        "opentelemetry-util-http==0.27b0",
        "deprecated==1.2.12",
        "google>=3.0.0",
        "pyyaml",
        "protobuf<=3.20.1"
    ],
    entry_points = {
        'console_scripts': [
            'tyke-instrument = tyke.agent.autoinstrumentation.tyke_instrument:run',
        ],
    }
)



'''constants'''
from tyke import version

EXCEPTION_MESSAGE = 'Failed to initialize %s instrumentation wrapper: exception=%s, stacktrace=%s'
INST_WRAP_EXCEPTION_MSSG = 'Failed to initialize %s instrumentation wrapper: exception=%s, stacktrace=%s' # pylint: disable=C0301
INST_RUNTIME_EXCEPTION_MSSG = 'An error occurred in %s: exception=%s, stacktrace=%s'
TELEMETRY_SDK_VERSION = version.__version__
TELEMETRY_SDK_NAME = 'tyke'
TELEMETRY_SDK_LANGUAGE = 'python'



'''Tyke wrapper around OTel GRPC instrumentor'''
import json
import logging
import traceback
import grpc
from google.protobuf.json_format import MessageToDict
from opentelemetry import trace
from opentelemetry.instrumentation.grpc import (
    GrpcInstrumentorServer,
    GrpcInstrumentorClient,
    _server,
    _client,
)
from opentelemetry.instrumentation.grpc.version import __version__
from opentelemetry.instrumentation.grpc.grpcext import intercept_channel
from wrapt import wrap_function_wrapper as _wrap # pylint: disable=R0801
from tyke.agent import constants
from tyke.agent.filter.registry import Registry, TYPE_RPC
from tyke.agent.instrumentation import BaseInstrumentorWrapper

# Initialize logger with local module name
logger = logging.getLogger(__name__) # pylint: disable=C0103

# The main entry point for a wrapper around the OTel grpc:server instrumentation module
class GrpcInstrumentorServerWrapper(GrpcInstrumentorServer, BaseInstrumentorWrapper):
    '''Tyke wrapper around OTel grpc:server instrumentor class'''
    # Construtor
    def __init__(self):
        '''constructor'''
        logger.debug('Entering GrpcInstrumentorServerWrapper constructor.')
        super().__init__()
        self._original_wrapper_func = None

    # Enable wrapper instrumentation
    def instrument(self, **kwargs) -> None:
        '''instrument grpc:server'''
        logger.debug('Entering GrpcInstrumentorServerWrapper.instument().')
        super().instrument()

    # Remove wrapper instrumentation
    def uninstrument(self, **kwargs) -> None:
        '''disable grpc:server instrumentation.'''
        logger.debug('Entering GrpcInstrumentorServerWrapper.uninstrument()')
        super().uninstrument()

    # Internal enable wrapper instrumentation
    def _instrument(self, **kwargs) -> None:
        '''Enable wrapper instrumentation internal call'''
        logger.debug('Entering GrpcInstrumentorServerWrapper._instument().')
        super()._instrument(**kwargs)
        self._original_wrapper_func = grpc.server

        def server_wrapper(*args, **kwargs) -> None:
            logger.debug('Entering wrapper interceptors set')
            logger.debug(
                'Setting server_interceptor_wrapper() as interceptor.')
            kwargs["interceptors"] = [server_interceptor_wrapper(self)]
            return self._original_wrapper_func(*args, **kwargs)
        grpc.server = server_wrapper

    # Internal disable wrapper instrumentation
    def _uninstrument(self, **kwargs) -> None:
        '''Disable wrapper instrumentation internal call'''
        logger.debug('Entering GrpcInstrumentorServerWrapper._uninstrument()')
        super()._uninstrument(**kwargs)

# The main entry point for a wrapper around the OTel grpc:client instrumentation module
class GrpcInstrumentorClientWrapper(GrpcInstrumentorClient, BaseInstrumentorWrapper):
    '''Tyke wrapper around OTel grpc:client instrumentor class'''
    # constructor
    def __init__(self):
        '''constructor'''
        logger.debug('Entering GrpcInstrumentorClientWrapper constructor.')
        super().__init__()

    # Internal initialize instrumentation
    def _instrument(self, **kwargs) -> None:
        '''Internal initialize instrumentation'''
        logger.debug('Entering GrpcInstrumentorClientWrapper._instrument().')
        super()._instrument(**kwargs)
# Uncomment this to reenable the client handler.
#    for ctype in self._which_channel(kwargs):
#      _wrap(
#        "grpc", ctype, self.wrapper_fn_wrapper,
#    )

    # Internal disable instrumentation
    def _uninstrument(self, **kwargs) -> None:
        '''Internal disable instrumentation'''
        logger.debug('Entering GrpcInstrumentorClientWrapper._uninstrument().')
        super()._uninstrument(**kwargs)

    # Wrap function for initializing the the request handler
    def wrapper_fn_wrapper(self, original_func, instance, args, kwargs) -> None: # pylint: disable=W0613,R0201
        '''Wrap function for initializing the the request handler'''
        channel = original_func(*args, **kwargs)
        tracer_provider = kwargs.get("tracer_provider")
        return intercept_channel(
            channel, client_interceptor_wrapper(
                tracer_provider=tracer_provider),
        )

# Initialize the server handler
def server_interceptor_wrapper(gisw, tracer_provider=None) -> None:
    '''Helper function to set interceptor.'''
    logger.debug('Entering server_interceptor_wrapper().')
    tracer = trace.get_tracer(__name__, __version__, tracer_provider)
    return OpenTelemetryServerInterceptorWrapper(tracer, gisw)

# Initialize the client handler
def client_interceptor_wrapper(tracer_provider) -> None:
    '''Helper function to set interceptor.'''
    logger.debug('Entering client_interceptor_wrapper().')
    tracer = trace.get_tracer(__name__, __version__, tracer_provider)
    return OpenTelemetryClientInterceptorWrapper(tracer)

# Wrapper around Server-side telemetry context
class _OpenTelemetryWrapperServicerContext(_server._OpenTelemetryServicerContext): # pylint: disable=W0212,W0223
    '''grpc:server telemetry context'''
    def __init__(self, servicer_context, active_span):
        '''constructor'''
        logger.debug(
            'Entering _OpenTelemetryWrapperServicerContext.__init__().')
        super().__init__(servicer_context, active_span)
        self._response_headers = ()

    def set_trailing_metadata(self, *args, **kwargs) -> None:
        """Override trailing metadata(response headers) method.
        Allows us to capture the response headers"""
        logger.debug(
            'Entering _OpenTelemetryWrapperServicerContext.set_trailing_metadata().')
        self._response_headers = args
        return self._servicer_context.set_trailing_metadata(*args, **kwargs)

    def get_trailing_metadata(self) -> tuple:
        '''Return response headers'''
        return self._response_headers

# Wrapper around server-side interceptor


class OpenTelemetryServerInterceptorWrapper(_server.OpenTelemetryServerInterceptor): # pylint: disable=R0903
    '''Tyke wrapper around OTel grpc:server Instrumentor class'''
    def __init__(self, tracer, gisw):
        '''constructor'''
        logger.debug(
            'Entering OpenTelemetryServerInterceptorWrapper.__init__().')
        super().__init__(tracer)
        self._gisw = gisw

    def intercept_service(self, continuation, handler_call_details):
        '''Setup interceptor'''
        logger.debug(
            'Entering OpenTelemetryServerInterceptorWrapper.intercept_service().')

        def telemetry_wrapper(behavior, request_streaming, response_streaming): # pylint: disable=W0613
            '''Setup interceptor helper for unary requests.'''
            logger.debug(
                'Entering OpenTelemetryServerInterceptorWrapper.telemetry_wrapper().')

            def telemetry_interceptor(request_or_iterator, context) -> None:
                '''Process request for tyke.'''
                logger.debug(
                    'Entering OpenTelemetryServerInterceptorWrapper.telemetry_interceptor().')
                # handle streaming responses specially
                if response_streaming:
                    return self._intercept_server_stream(
                        behavior,
                        handler_call_details,
                        request_or_iterator,
                        context,
                    )

                span = context._active_span # pylint: disable=W0212

                invocation_metadata = dict(handler_call_details.invocation_metadata)
                req_dict = MessageToDict(request_or_iterator)
                self._gisw.generic_rpc_request_handler(
                    invocation_metadata, json.dumps(req_dict), span)
                try:
                    block_result = Registry().apply_filters(span,
                                                            '',
                                                            invocation_metadata,
                                                            request_or_iterator,
                                                            TYPE_RPC)
                    if block_result:
                        logger.debug('should block evaluated to true, aborting with 403')
                        return context.abort(grpc.StatusCode.PERMISSION_DENIED, 'Permission Denied')

                    # Capture response
                    context = _OpenTelemetryWrapperServicerContext(
                        context, span)
                    response = behavior(request_or_iterator, context)
                    trailing_metadata = context.get_trailing_metadata()
                    if len(trailing_metadata) > 0:
                        trailing_metadata = dict(trailing_metadata[0])
                    else:
                        trailing_metadata = {}

                    response_dict = MessageToDict(response)
                    self._gisw.generic_rpc_response_handler(
                        trailing_metadata, json.dumps(response_dict), span)

                    return response
                except Exception as error: # pylint: disable=W0703
                    # Bare exceptions are likely to be gRPC aborts, which
                    # we handle in our context wrapper.
                    # Here, we're interested in uncaught exceptions.
                    # pylint:disable=unidiomatic-typecheck
                    if type(error) != Exception:
                        span.record_exception(error)
                        raise error
                    return None
            return telemetry_interceptor
        return _server._wrap_rpc_behavior(continuation(handler_call_details), telemetry_wrapper) # pylint: disable=W0212

    def _intercept_server_stream(
            self,
            behavior,
            handler_call_details,
            request_or_iterator,
            context) -> None:
        '''Setup interceptor helper for streaming requests.'''
        logger.debug(
            'Entering OpenTelemetryServerInterceptorWrapper.intercept_server_stream().')
        # COME_BACK -- need to implement this

# Wrapper around client-side interceptor
class OpenTelemetryClientInterceptorWrapper(_client.OpenTelemetryClientInterceptor):
    '''Tyke wrapper around OTel grpc:client instrumentor class.'''
    def __init__(self, tracer):
        '''constructor'''
        logger.debug(
            'Entering OpenTelemetryClientInterceptorWrapper.__init__().')
        super().__init__(tracer)

    def intercept_unary(self, request, metadata, client_info, invoker) -> None:
        '''Process unary request for tyke.'''
        logger.debug(
            'Entering OpenTelemetryClientInterceptorWrapper.intercept_unary().')
        try:
            # Not sure how to obtain span object here
            result = invoker(request, metadata)  # pylint:disable=W0612
            # Not sure how to obtain trailing metadata here
        except grpc.RpcError as err:
            logger.error(constants.INST_RUNTIME_EXCEPTION_MSSG,
                         'processing client request',
                         err,
                         traceback.format_exc())
            raise err

#        return self._trace_result(result)

    def intercept_stream(
            self,
            request_or_iterator,
            metadata,
            client_info,
            invoker
    ) -> None:
        '''process streaming request for tyke'''
        logger.debug(
            'Entering OpenTelemetryClientInterceptorWrapper.intercept_stream().')
        # COME_BACK -- need to implement this



'''The default configuration for pythonagent'''
DEFAULT_AGENT_CONFIG = {
    '_use_console_span_exporter': False,
    'enabled': True,
    'propagation_formats': ['TRACECONTEXT'],
    'service_name': 'pythonagent',
    'reporting': {
        'endpoint': 'http://localhost:4317',
        'secure': False,
        'trace_reporter_type': 'OTLP',
        'token': '',
    },
    'data_capture': {
        'http_headers': {
            'request': True,
            'response': True,
        },
        'http_body': {
            'request': True,
            'response': True,
        },
        'rpc_metadata': {
            'request': True,
            'response': True,
        },
        'rpc_body': {
            'request': True,
            'response': True,
        },
        'body_max_size_bytes': 131072
    },
    'resource_attributes': {}
}



'''Tyke wrapper around OTel instrumentation class'''
import logging
from opentelemetry.instrumentation.requests import RequestsInstrumentor
from tyke.agent.instrumentation import BaseInstrumentorWrapper

# Initialize logger with local module name
logger = logging.getLogger(__name__) # pylint: disable=C0103

def get_active_span_for_call_wrapper(requests_wrapper):
    '''Helper function to setup call wrapper function'''
    logger.debug('Entering get_active_span_for_call_wrapper().')

    def get_active_span_for_call(span, response) -> None:
        '''Tyke call wrapper function'''
        logger.debug('Entering get_active_span_for_request().')
        response_content = None
        if hasattr(response, 'content'):
            logger.debug('Converting response message body to string.')
            response_content = response.content.decode()
        else:
            logger.debug('No response message body. Setting to blank string.')
            response_content = ''
        request_content = None
        if hasattr(response.request, 'content'):
            logger.debug('Converting request message body to string.')
            request_content = response.request.content.decode()
        else:
            logger.debug('No request message body. Setting to blank string.')
            request_content = ''

        if span.is_recording():
            logger.debug('Span is recording.')
            requests_wrapper.generic_request_handler(
                response.request.headers, request_content, span)
            requests_wrapper.generic_response_handler(
                response.headers, response_content, span)
    return get_active_span_for_call

def tyke_name_callback(method, url) -> str:
    '''generate span name'''
    logger.debug('Entering tyke_name_callback(), method=%s, url=%s.', method, url)
    return method + ' ' + url

class RequestsInstrumentorWrapper(RequestsInstrumentor, BaseInstrumentorWrapper):
    '''Tyke wrapper around OTel requests instrumentor class'''
    # Constructor
    def __init__(self):
        logger.debug('Entering RequestsInstrumentorWrapper.__init__().')
        super().__init__()

    def _instrument(self, **kwargs) -> None:
        '''internal enable instrumentation'''
        super()._instrument(
            tracer_provider=kwargs.get("tracer_provider"),
            span_callback=get_active_span_for_call_wrapper(self),
            name_callback=tyke_name_callback
        )

    def _uninstrument(self, **kwargs) -> None:
        '''internal disable instrumentation'''
        super()._uninstrument()
